#!/usr/bin/env python

"""
Test the  WBEMConnection iteration methods (Iter...) by mocking the lower level
WBEMConnection methods that are called by the iteration methods.

Thus, to mock iterEnumerateInstances, we must mock OpenEnumerateInstances,
pullInstanceseWithPath, CloseEnumeration, and EnumerateInstances to be complete.
With that we can execute the variations on IterEnumerateInstances.

This test uses pytest as the test tool.

It specifically tests:
1. To assure that use_pull_operations works
2. The correct parameters are passed to the lower level methods.
3. The correct results are returned.

It does not test for valid parameters if the parameters are tested by lower
level methods because those are the methods that are mocked.
"""

from __future__ import absolute_import, print_function

import pytest
import six

from mock import Mock

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import WBEMConnection, CIMInstance, CIMClass, CIMInstanceName, \
    CIMClassName, CIMProperty, CIMError, CIM_ERR_NOT_SUPPORTED  # noqa: E402
from pywbem.config import DEFAULT_ITER_MAXOBJECTCOUNT  # noqa: E402
from pywbem._cim_operations import pull_inst_result_tuple, \
    pull_path_result_tuple, pull_query_result_tuple  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


@pytest.fixture
def tst_class():
    """
    Pytest fixture to return a class.
    """
    cls = CIMClass(
        'CIM_Foo',
        properties=[
            CIMProperty('InstanceID', None, type='string'),
        ],
    )
    return cls


@pytest.fixture
def tst_insts():
    """
    Pytest fixture to return N instances of the class returned by tst_class.
    """
    obj_count = 2
    rtn = []
    for i in six.moves.range(obj_count):
        instanceid = str(i + 1000)
        obj = CIMInstance(
            'CIM_Foo',
            properties=[
                CIMProperty('InstanceID', instanceid, type='string'),
                CIMProperty('Chicken', 'Ham', type='string'),
            ],
            path=CIMInstanceName(
                'CIM_Foo',
                keybindings={'InstanceID': instanceid},
                host='woot.com',
                namespace='root/cimv2',
            ),
        )
        rtn.append(obj)
    return rtn


@pytest.fixture
def tst_paths():
    """
    Pytest fixture to return N instance paths of the class returned by
    tst_class.
    """
    obj_count = 2
    rtn = []
    for i in six.moves.range(obj_count):
        instanceid = str(i + 1000)
        obj = CIMInstanceName(
            'CIM_Foo',
            keybindings={'InstanceID': instanceid},
            host='woot.com',
            namespace='root/cimv2',
        )
        rtn.append(obj)
    return rtn


########################################################################
#
#               IterEnumerateInstances tests
#
########################################################################

class TestIterEnumerateInstances(object):
    """Test IterEnumerateInstances execution"""

    @pytest.mark.parametrize(
        "use_pull_param", [None, False]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "cim_classname", [False, True]
    )
    @pytest.mark.parametrize(
        "result_no_host_ns", [False, True]
    )
    @pytest.mark.parametrize(
        "lo, di, iq, ico", [[None, None, None, None],
                            [False, False, False, False],
                            [True, True, True, True]]
    )
    @pytest.mark.parametrize(
        "pl", [None, 'pl1', ['pl1', 'pl2']]
    )
    def test_trad_operation_success(self, use_pull_param, tst_insts, ns,
                                    cim_classname, result_no_host_ns,
                                    di, iq, lo, ico, pl,):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterEnumerateInstances using the traditional operations,
        either by requesting it via use_pull_operations=False, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the Open operation. This must succeed.

        The test is parameterized to test variations on all parameters. Note
        that it only tests legal parameters.

        The test confirms that the traditional operation receives the correct
        values for all parameters.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        if result_no_host_ns:
            rtn_insts = []
            for i in tst_insts:
                rtn_inst = i.copy()
                rtn_inst.path.host = None
                rtn_inst.path.namespace = None
                rtn_insts.append(rtn_inst)
                i.path.namespace = ns or conn.default_namespace
                i.path.host = conn.host
        else:
            rtn_insts = list(tst_insts)

        # mock original function works, open returns CIMError
        conn.EnumerateInstances = Mock(return_value=rtn_insts)
        conn.OpenEnumerateInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_enum_inst_pull_operations == use_pull_param

        if cim_classname:
            classname = CIMClassName('CIM_Foo', namespace=ns)
        else:
            classname = 'CIM_Foo'

        result = list(conn.IterEnumerateInstances(
            classname,
            namespace=ns,
            LocalOnly=lo,
            DeepInheritance=di,
            IncludeQualifiers=iq,
            IncludeClassOrigin=ico,
            PropertyList=pl))

        conn.EnumerateInstances.assert_called_with(
            classname,
            namespace=ns,
            LocalOnly=lo,
            DeepInheritance=di,
            IncludeQualifiers=iq,
            IncludeClassOrigin=ico,
            PropertyList=pl)

        assert conn._use_enum_inst_pull_operations is False
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_insts

    def test_operation_fail(self, tst_insts):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterEnumerateInstances forcing the use of pull operations
        via use_pull_operations=True, and the Open operation raises
        NOT_SUPPORTED. This must fail.
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.EnumerateInstances = Mock(return_value=tst_insts)
        conn.OpenEnumerateInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'description'))

        assert conn.use_pull_operations is True
        # pylint: disable=protected-access
        assert conn._use_enum_inst_pull_operations is True

        with pytest.raises(CIMError) as exec_info:
            _ = list(conn.IterEnumerateInstances('CIM_Foo'))

        exc = exec_info.value
        assert exc.status_code_name == 'CIM_ERR_NOT_SUPPORTED'

        # pylint: disable=protected-access
        assert conn._use_enum_inst_pull_operations is True
        assert conn.use_pull_operations is True

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "lo, di, iq, ico", [[None, None, None, None],
                            [False, False, False, False],
                            [True, True, True, True]]
    )
    @pytest.mark.parametrize(
        "pl", [None, 'pl1', ['pl1', 'pl2']]
    )
    @pytest.mark.parametrize(
        "fl,fq", [(None, None), ('FQL', 'SELECT')]
    )
    @pytest.mark.parametrize(
        "ot, coe, moc", [[None, None, 100],
                         [10, False, 100]]
    )
    def test_open_operation_success(self, use_pull_param, tst_insts, ns,
                                    lo, di, iq, ico, pl,
                                    fl, fq, ot, coe, moc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterEnumerateInstances using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation. This must succeed.

        The test is parameterized to test variations on all parameters. Note
        that it only tests legal parameters.

        The test confirms that the Open operation receives the correct
        values for all parameters, and that the Close operation is not called.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.EnumerateInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenEnumerateInstances = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts,
                                                     eos=True,
                                                     context=None))
        conn.CloseEnumeration = Mock()

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_enum_inst_pull_operations == use_pull_param

        result = list(conn.IterEnumerateInstances(
            'CIM_Foo',
            namespace=ns,
            LocalOnly=lo,
            DeepInheritance=di,
            IncludeQualifiers=iq,
            IncludeClassOrigin=ico,
            PropertyList=pl,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc))

        conn.OpenEnumerateInstances.assert_called_with(
            'CIM_Foo',
            namespace=ns,
            DeepInheritance=di,
            IncludeClassOrigin=ico,
            PropertyList=pl,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)

        conn.CloseEnumeration.assert_not_called()

        # pylint: disable=protected-access
        assert conn._use_enum_inst_pull_operations is True
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_insts

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    def test_pull_operation_success(self, tst_insts, use_pull_param):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterEnumerateInstances using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation. This must succeed.

        The Open operation returns no result objects, and the Pull operation
        returns all result objects and EOS.

        This does not retest the variations of input parameters because it
        is only confirming that the Pull operation executes.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        ctx = ('blah', conn.default_namespace)

        conn.EnumerateInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenEnumerateInstances = \
            Mock(return_value=pull_inst_result_tuple(instances=[], eos=False,
                                                     context=ctx))
        conn.PullInstancesWithPath = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts,
                                                     eos=True,
                                                     context=None))

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_enum_inst_pull_operations == use_pull_param

        result = list(conn.IterEnumerateInstances('CIM_Foo'))

        assert conn._use_enum_inst_pull_operations is True
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_insts

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "max_obj_count", [1, 2]
    )
    @pytest.mark.parametrize(
        "close_after", [0, 1, 2]
    )
    def test_closed_operation_success(self, tst_insts, use_pull_param,
                                      max_obj_count, close_after):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterEnumerateInstances using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation.

        The iteration is closed after a number of retrievals, sometimes
        prematurely and sometimes not.

        This does not retest the variations of input parameters because it
        is only confirming that the Close operation executes if needed.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        tst_insts_open = tst_insts[0:max_obj_count]
        tst_insts_pull = tst_insts[max_obj_count:]
        ctx = ('blah', conn.default_namespace)

        conn.EnumerateInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenEnumerateInstances = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts_open,
                                                     eos=False,
                                                     context=ctx))
        conn.PullInstancesWithPath = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts_pull,
                                                     eos=True,
                                                     context=None))
        conn.CloseEnumeration = Mock()

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_enum_inst_pull_operations == use_pull_param

        gen = conn.IterEnumerateInstances(
            'CIM_Foo', MaxObjectCount=max_obj_count)
        result = []
        for _ in six.moves.range(close_after):
            obj = next(gen)
            result.append(obj)
        gen.close()  # close prematurely

        if close_after == 0 or close_after > max_obj_count:
            # Pull has been called; enumeration is exhausted; enumeration was
            # closed automatically within Pull.
            conn.CloseEnumeration.assert_not_called()
        else:
            # Pull has not been called; enumeration is not exhausted;
            # enumeration was closed by calling Close.
            conn.CloseEnumeration.assert_called_with(ctx)
            assert conn._use_enum_inst_pull_operations is True

        assert conn.use_pull_operations == use_pull_param
        assert result == tst_insts[0:close_after]

    @pytest.mark.parametrize(
        "kwargs, exp_exc",
        [
            [dict(OperationTimeout='bla'), TypeError()],
            [dict(OperationTimeout=-1), ValueError()],
            [dict(MaxObjectCount=0), ValueError()],
            [dict(MaxObjectCount=-1), ValueError()],
            [dict(MaxObjectCount=None), ValueError()],
            [dict(MaxObjectCount='bla'), TypeError()],
            [dict(MaxObjectCount='bla'), TypeError()],
        ]
    )
    def test_open_invalid_params(self, tst_insts, kwargs, exp_exc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterEnumerateInstances using the pull operations,
        with invalid values and types of those input parameters that are
        actually checked (i.e. OperationTimeout and MaxObjectCount).
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.EnumerateInstances = Mock(return_value=tst_insts)
        conn.OpenEnumerateInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        with pytest.raises(type(exp_exc)):
            _ = list(conn.IterEnumerateInstances(
                'CIM_Foo',
                **kwargs))

    @pytest.mark.parametrize(
        "kwargs, exp_exc",
        [
            [dict(FilterQueryLanguage='CQL'), ValueError()],
            [dict(FilterQuery='Prop=42'), ValueError()],
            [dict(ContinueOnError=True), ValueError()],
        ]
    )
    def test_trad_invalid_params(self, tst_insts, kwargs, exp_exc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterEnumerateInstances using the traditional operations,
        with invalid values and types of those input parameters that are
        actually checked (i.e. FilterQueryLanguage, FilterQuery,
        ContinueOnError).
        """
        conn = WBEMConnection('dummy', use_pull_operations=False)

        conn.EnumerateInstances = Mock(return_value=tst_insts)
        conn.OpenEnumerateInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        with pytest.raises(type(exp_exc)):
            _ = list(conn.IterEnumerateInstances(
                'CIM_Foo',
                **kwargs))


########################################################################
#
#               IterEnumerateInstancePaths tests
#
########################################################################

class TestIterEnumerateInstancePaths(object):
    """Test IterEnumerateInstancePaths execution"""

    @pytest.mark.parametrize(
        "use_pull_param", [None, False]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "cim_classname", [False, True]
    )
    @pytest.mark.parametrize(
        "result_no_host_ns", [False, True]
    )
    def test_trad_operation_success(self, use_pull_param, tst_paths, ns,
                                    cim_classname, result_no_host_ns):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterEnumerateInstancePaths using the traditional operations,
        either by requesting it via use_pull_operations=False, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the Open operation. This must succeed.

        The test is parameterized to test variations on all parameters. Note
        that it only tests legal parameters.

        The test confirms that the traditional operation receives the correct
        values for all parameters.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        if result_no_host_ns:
            rtn_paths = []
            for p in tst_paths:
                rtn_path = p.copy()
                rtn_path.host = None
                rtn_path.namespace = None
                rtn_paths.append(rtn_path)
                p.namespace = ns or conn.default_namespace
                p.host = conn.host
        else:
            rtn_paths = tst_paths

        conn.EnumerateInstanceNames = Mock(return_value=rtn_paths)
        conn.OpenEnumerateInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_enum_path_pull_operations == use_pull_param

        if cim_classname:
            classname = CIMClassName('CIM_Foo', namespace=ns)
        else:
            classname = 'CIM_Foo'

        result = list(conn.IterEnumerateInstancePaths(
            classname,
            namespace=ns))

        conn.EnumerateInstanceNames.assert_called_with(
            classname,
            namespace=ns)

        # pylint: disable=protected-access
        assert conn._use_enum_path_pull_operations is False
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_paths

    def test_operation_fail(self, tst_insts):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterEnumerateInstancePaths forcing the use of pull operations
        via use_pull_operations=True, and the Open operation raises
        NOT_SUPPORTED. This must fail.
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.EnumerateInstanceNames = Mock(return_value=tst_insts)
        conn.OpenEnumerateInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'description'))

        assert conn.use_pull_operations is True
        # pylint: disable=protected-access
        assert conn._use_enum_path_pull_operations is True

        with pytest.raises(CIMError) as exec_info:
            _ = list(conn.IterEnumerateInstancePaths('CIM_Foo'))

        exc = exec_info.value
        assert exc.status_code_name == 'CIM_ERR_NOT_SUPPORTED'

        # pylint: disable=protected-access
        assert conn._use_enum_path_pull_operations is True
        assert conn.use_pull_operations is True

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "fl,fq", [(None, None), ('FQL', 'SELECT')]
    )
    @pytest.mark.parametrize(
        "ot, coe, moc", [[None, None, 100],
                         [10, False, 100]]
    )
    def test_open_operation_success(self, tst_paths, ns, fl, fq, ot,
                                    coe, moc, use_pull_param):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterEnumerateInstancePaths using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation. This must succeed.

        The test is parameterized to test variations on all parameters. Note
        that it only tests legal parameters.

        The test confirms that the Open operation receives the correct
        values for all parameters, and that the Close operation is not called.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.EnumerateInstanceNames = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenEnumerateInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths,
                                                     eos=True,
                                                     context=None))
        conn.CloseEnumeration = Mock()

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_enum_path_pull_operations == use_pull_param

        result = list(conn.IterEnumerateInstancePaths(
            'CIM_Foo',
            namespace=ns,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc))

        conn.OpenEnumerateInstancePaths.assert_called_with(
            'CIM_Foo',
            namespace=ns,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)

        conn.CloseEnumeration.assert_not_called()

        # pylint: disable=protected-access
        assert conn._use_enum_path_pull_operations is True
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_paths

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    def test_pull_operation_success(self, tst_paths, use_pull_param):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterEnumerateInstancePaths using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation. This must succeed.

        The Open operation returns no result objects, and the Pull operation
        returns all result objects and EOS.

        This does not retest the variations of input parameters because it
        is only confirming that the Pull operation executes.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        ctx = ('blah', conn.default_namespace)

        conn.EnumerateInstanceNames = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenEnumerateInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=[], eos=False,
                                                     context=ctx))
        conn.PullInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths,
                                                     eos=True,
                                                     context=None))

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_enum_path_pull_operations == use_pull_param

        result = list(conn.IterEnumerateInstancePaths('CIM_Foo'))

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_enum_path_pull_operations is True

        assert result == tst_paths

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "max_obj_count", [1, 2]
    )
    @pytest.mark.parametrize(
        "close_after", [0, 1, 2]
    )
    def test_closed_operation_success(self, tst_paths, use_pull_param,
                                      max_obj_count, close_after):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterEnumerateInstancePaths using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation.

        The iteration is closed after a number of retrievals, sometimes
        prematurely and sometimes not.

        This does not retest the variations of input parameters because it
        is only confirming that the Close operation executes if needed.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        tst_paths_open = tst_paths[0:max_obj_count]
        tst_paths_pull = tst_paths[max_obj_count:]
        ctx = ('blah', conn.default_namespace)

        conn.EnumerateInstanceNames = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenEnumerateInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths_open,
                                                     eos=False,
                                                     context=ctx))
        conn.PullInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths_pull,
                                                     eos=True,
                                                     context=None))
        conn.CloseEnumeration = Mock()

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_enum_path_pull_operations == use_pull_param

        gen = conn.IterEnumerateInstancePaths(
            'CIM_Foo', MaxObjectCount=max_obj_count)
        result = []
        for _ in six.moves.range(close_after):
            obj = next(gen)
            result.append(obj)
        gen.close()  # close prematurely

        if close_after == 0 or close_after > max_obj_count:
            # Pull has been called; enumeration is exhausted; enumeration was
            # closed automatically within Pull.
            conn.CloseEnumeration.assert_not_called()
        else:
            # Pull has not been called; enumeration is not exhausted;
            # enumeration was closed by calling Close.
            conn.CloseEnumeration.assert_called_with(ctx)
            assert conn._use_enum_path_pull_operations is True

        assert conn.use_pull_operations == use_pull_param
        assert result == tst_paths[0:close_after]

    @pytest.mark.parametrize(
        "kwargs, exp_exc",
        [
            [dict(OperationTimeout='bla'), TypeError()],
            [dict(OperationTimeout=-1), ValueError()],
            [dict(MaxObjectCount=0), ValueError()],
            [dict(MaxObjectCount=-1), ValueError()],
            [dict(MaxObjectCount=None), ValueError()],
            [dict(MaxObjectCount='bla'), TypeError()],
            [dict(MaxObjectCount='bla'), TypeError()],
        ]
    )
    def test_open_invalid_params(self, tst_paths, kwargs, exp_exc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterEnumerateInstancePaths using the pull operations,
        with invalid values and types of those input parameters that are
        actually checked (i.e. OperationTimeout and MaxObjectCount).
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.EnumerateInstanceNames = Mock(return_value=tst_paths)
        conn.OpenEnumeratePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        with pytest.raises(type(exp_exc)):
            _ = list(conn.IterEnumerateInstancePaths(
                'CIM_Foo',
                **kwargs))

    @pytest.mark.parametrize(
        "kwargs, exp_exc",
        [
            [dict(FilterQueryLanguage='CQL'), ValueError()],
            [dict(FilterQuery='Prop=42'), ValueError()],
            [dict(ContinueOnError=True), ValueError()],
        ]
    )
    def test_trad_invalid_params(self, tst_paths, kwargs, exp_exc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterEnumerateInstancePaths using the traditional operations,
        with invalid values and types of those input parameters that are
        actually checked (i.e. FilterQueryLanguage, FilterQuery,
        ContinueOnError).
        """
        conn = WBEMConnection('dummy', use_pull_operations=False)

        conn.EnumerateInstanceNames = Mock(return_value=tst_paths)
        conn.OpenEnumeratePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        with pytest.raises(type(exp_exc)):
            _ = list(conn.IterEnumerateInstancePaths(
                'CIM_Foo',
                **kwargs))


########################################################################
#
#               IterReferenceInstances tests
#
########################################################################

class TestIterReferenceInstances(object):
    """Test IterReferenceInstances execution"""

    @staticmethod
    def target_path(ns):
        """
        Return a valid target path for the calls
        """
        return CIMInstanceName('CIM_blah',
                               keybindings={'InstanceID': "1234"},
                               host='woot.com',
                               namespace=ns)

    @pytest.mark.parametrize(
        "use_pull_param", [None, False]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "iq, ico", [[None, None],
                    [False, False],
                    [True, True]]
    )
    # result_role, result_class
    @pytest.mark.parametrize(
        "ro, rc", [[None, None],
                   [None, 'CIM_Blah'],
                   ['RRole', None],
                   ['RRole', 'CIM_Blah']]
    )
    @pytest.mark.parametrize(
        "pl", [None, 'pl1', ['pl1', 'pl2']]
    )
    def test_trad_operation_success(self, use_pull_param, tst_insts, ns,
                                    rc, ro, iq, ico, pl,):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterReferenceInstances using the traditional operations,
        either by requesting it via use_pull_operations=False, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the Open operation. This must succeed.

        The test is parameterized to test variations on all parameters. Note
        that it only tests legal parameters.

        The test confirms that the traditional operation receives the correct
        values for all parameters.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        # mock original function works, open returns CIMError
        conn.References = Mock(return_value=tst_insts)
        conn.OpenReferenceInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_ref_inst_pull_operations == use_pull_param

        result = list(conn.IterReferenceInstances(
            self.target_path(ns),
            ResultClass=rc,
            Role=ro,
            IncludeQualifiers=iq,
            IncludeClassOrigin=ico,
            PropertyList=pl))

        conn.References.assert_called_with(
            self.target_path(ns),
            ResultClass=rc,
            Role=ro,
            IncludeQualifiers=iq,
            IncludeClassOrigin=ico,
            PropertyList=pl)

        assert conn._use_ref_inst_pull_operations is False
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_insts

    def test_operation_fail(self, tst_insts):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterReferenceInstances forcing the use of pull operations
        via use_pull_operations=True, and the Open operation raises
        NOT_SUPPORTED. This must fail.
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.References = Mock(return_value=tst_insts)
        conn.OpenReferenceInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'description'))

        assert conn.use_pull_operations is True
        # pylint: disable=protected-access
        assert conn._use_ref_inst_pull_operations is True

        with pytest.raises(CIMError) as exec_info:
            _ = list(conn.IterReferenceInstances(
                self.target_path('root/cimv2')))

        exc = exec_info.value
        assert exc.status_code_name == 'CIM_ERR_NOT_SUPPORTED'

        # pylint: disable=protected-access
        assert conn._use_ref_inst_pull_operations is True
        assert conn.use_pull_operations is True

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "iq, ico", [[None, None],
                    [False, False],
                    [True, True]]
    )
    # result_role, result_class
    @pytest.mark.parametrize(
        "ro, rc", [[None, None],
                   [None, 'CIM_Blah'],
                   ['RRole', None],
                   ['RRole', 'CIM_Blah']]
    )
    @pytest.mark.parametrize(
        "pl", [None, 'pl1', ['pl1', 'pl2']]
    )
    @pytest.mark.parametrize(
        "fl,fq", [(None, None), ('FQL', 'SELECT')]
    )
    @pytest.mark.parametrize(
        "ot, coe, moc", [[None, None, 100],
                         [10, False, 100]]
    )
    def test_open_operation_success(self, use_pull_param, tst_insts, ns,
                                    rc, ro, iq, ico, pl, fl, fq, ot, coe,
                                    moc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterReferenceInstances using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation. This must succeed.

        The test is parameterized to test variations on all parameters. Note
        that it only tests legal parameters.

        The test confirms that the Open operation receives the correct
        values for all parameters, and that the Close operation is not called.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.References = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenReferenceInstances = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts,
                                                     eos=True,
                                                     context=None))
        conn.CloseEnumeration = Mock()

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_ref_inst_pull_operations == use_pull_param

        result = list(conn.IterReferenceInstances(
            self.target_path(ns),
            ResultClass=rc,
            Role=ro,
            IncludeQualifiers=iq,
            IncludeClassOrigin=ico,
            PropertyList=pl,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc))

        conn.OpenReferenceInstances.assert_called_with(
            self.target_path(ns),
            ResultClass=rc,
            Role=ro,
            IncludeClassOrigin=ico,
            PropertyList=pl,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)

        conn.CloseEnumeration.assert_not_called()

        # pylint: disable=protected-access
        assert conn._use_ref_inst_pull_operations is True
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_insts

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    def test_pull_operation_success(self, tst_insts, use_pull_param):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterReferenceInstances using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation. This must succeed.

        The Open operation returns no result objects, and the Pull operation
        returns all result objects and EOS.

        This does not retest the variations of input parameters because it
        is only confirming that the Pull operation executes.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        ctx = ('blah', conn.default_namespace)

        conn.References = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenReferenceInstances = \
            Mock(return_value=pull_inst_result_tuple(instances=[], eos=False,
                                                     context=ctx))
        conn.PullInstancesWithPath = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts,
                                                     eos=True,
                                                     context=None))

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_ref_inst_pull_operations == use_pull_param

        result = list(conn.IterReferenceInstances(
            self.target_path('root/cimv2')))

        assert conn._use_ref_inst_pull_operations is True
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_insts

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "max_obj_count", [1, 2]
    )
    @pytest.mark.parametrize(
        "close_after", [0, 1, 2]
    )
    def test_closed_operation_success(self, tst_insts, use_pull_param,
                                      max_obj_count, close_after):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterReferenceInstances using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation.

        The iteration is closed after a number of retrievals, sometimes
        prematurely and sometimes not.

        This does not retest the variations of input parameters because it
        is only confirming that the Close operation executes if needed.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        tst_insts_open = tst_insts[0:max_obj_count]
        tst_insts_pull = tst_insts[max_obj_count:]
        ctx = ('blah', conn.default_namespace)

        conn.References = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenReferenceInstances = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts_open,
                                                     eos=False,
                                                     context=ctx))
        conn.PullInstancesWithPath = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts_pull,
                                                     eos=True,
                                                     context=None))
        conn.CloseEnumeration = Mock()

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_ref_inst_pull_operations == use_pull_param

        gen = conn.IterReferenceInstances(
            self.target_path('root/cimv2'), MaxObjectCount=max_obj_count)
        result = []
        for _ in six.moves.range(close_after):
            obj = next(gen)
            result.append(obj)
        gen.close()  # close prematurely

        if close_after == 0 or close_after > max_obj_count:
            # Pull has been called; enumeration is exhausted; enumeration was
            # closed automatically within Pull.
            conn.CloseEnumeration.assert_not_called()
        else:
            # Pull has not been called; enumeration is not exhausted;
            # enumeration was closed by calling Close.
            conn.CloseEnumeration.assert_called_with(ctx)
            assert conn._use_ref_inst_pull_operations is True

        assert conn.use_pull_operations == use_pull_param
        assert result == tst_insts[0:close_after]

    @pytest.mark.parametrize(
        "kwargs, exp_exc",
        [
            [dict(OperationTimeout='bla'), TypeError()],
            [dict(OperationTimeout=-1), ValueError()],
            [dict(MaxObjectCount=0), ValueError()],
            [dict(MaxObjectCount=-1), ValueError()],
            [dict(MaxObjectCount=None), ValueError()],
            [dict(MaxObjectCount='bla'), TypeError()],
            [dict(MaxObjectCount='bla'), TypeError()],
        ]
    )
    def test_open_invalid_params(self, tst_insts, kwargs, exp_exc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterReferenceInstances using the pull operations,
        with invalid values and types of those input parameters that are
        actually checked (i.e. OperationTimeout and MaxObjectCount).
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.References = Mock(return_value=tst_insts)
        conn.OpenReferenceInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        with pytest.raises(type(exp_exc)):
            _ = list(conn.IterReferenceInstances(
                self.target_path('root/x'),
                **kwargs))

    @pytest.mark.parametrize(
        "kwargs, exp_exc",
        [
            [dict(FilterQueryLanguage='CQL'), ValueError()],
            [dict(FilterQuery='Prop=42'), ValueError()],
            [dict(ContinueOnError=True), ValueError()],
        ]
    )
    def test_trad_invalid_params(self, tst_insts, kwargs, exp_exc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterEnumerateInstances using the traditional operations,
        with invalid values and types of those input parameters that are
        actually checked (i.e. FilterQueryLanguage, FilterQuery,
        ContinueOnError).
        """
        conn = WBEMConnection('dummy', use_pull_operations=False)

        conn.References = Mock(return_value=tst_insts)
        conn.OpenReferenceInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        with pytest.raises(type(exp_exc)):
            _ = list(conn.IterReferenceInstances(
                self.target_path('root/x'),
                **kwargs))


########################################################################
#
#               IterReferenceInstancePaths tests
#
########################################################################

class TestIterReferenceInstancePaths(object):
    """Test IterReferenceInstancePaths execution"""

    @staticmethod
    def target_path(ns):
        """
        Return a valid target path for the calls
        """
        return CIMInstanceName('CIM_blah',
                               keybindings={'InstanceID': "1234"},
                               host='woot.com',
                               namespace=ns)

    @pytest.mark.parametrize(
        "use_pull_param", [None, False]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    # result_role, result_class
    @pytest.mark.parametrize(
        "ro, rc", [[None, None],
                   [None, 'CIM_Blah'],
                   ['RRole', None],
                   ['RRole', 'CIM_Blah']]
    )
    def test_trad_operation_success(self, use_pull_param, tst_insts, ns, rc,
                                    ro):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterReferenceInstancePaths using the traditional operations,
        either by requesting it via use_pull_operations=False, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the Open operation. This must succeed.

        The test is parameterized to test variations on all parameters. Note
        that it only tests legal parameters.

        The test confirms that the traditional operation receives the correct
        values for all parameters.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        # mock original function works, open returns CIMError
        conn.ReferenceNames = Mock(return_value=tst_insts)
        conn.OpenReferenceInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_ref_path_pull_operations == use_pull_param

        result = list(conn.IterReferenceInstancePaths(
            self.target_path(ns),
            ResultClass=rc,
            Role=ro))

        conn.ReferenceNames.assert_called_with(
            self.target_path(ns),
            ResultClass=rc,
            Role=ro)

        assert conn._use_ref_path_pull_operations is False
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_insts

    def test_operation_fail(self, tst_insts):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterReferenceInstancePaths forcing the use of pull operations
        via use_pull_operations=True, and the Open operation raises
        NOT_SUPPORTED. This must fail.
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.ReferenceNames = Mock(return_value=tst_insts)
        conn.OpenReferenceInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'description'))

        assert conn.use_pull_operations is True
        # pylint: disable=protected-access
        assert conn._use_ref_path_pull_operations is True

        with pytest.raises(CIMError) as exec_info:
            _ = list(conn.IterReferenceInstancePaths(
                self.target_path('root/cimv2')))

        exc = exec_info.value
        assert exc.status_code_name == 'CIM_ERR_NOT_SUPPORTED'

        # pylint: disable=protected-access
        assert conn._use_ref_path_pull_operations is True
        assert conn.use_pull_operations is True

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    # result_role, result_class
    @pytest.mark.parametrize(
        "ro, rc", [[None, None],
                   [None, 'CIM_Blah'],
                   ['RRole', None],
                   ['RRole', 'CIM_Blah']]
    )
    @pytest.mark.parametrize(
        "fl,fq", [(None, None), ('FQL', 'SELECT')]
    )
    @pytest.mark.parametrize(
        "ot, coe, moc", [[None, None, 100],
                         [10, False, 100]]
    )
    def test_open_operation_success(self, use_pull_param, tst_paths, ns,
                                    rc, ro, fl, fq, ot, coe, moc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterReferenceInstancePaths using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation. This must succeed.

        The test is parameterized to test variations on all parameters. Note
        that it only tests legal parameters.

        The test confirms that the Open operation receives the correct
        values for all parameters, and that the Close operation is not called.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.ReferenceNames = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenReferenceInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths,
                                                     eos=True,
                                                     context=None))
        conn.CloseEnumeration = Mock()

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_ref_path_pull_operations == use_pull_param

        result = list(conn.IterReferenceInstancePaths(
            self.target_path(ns),
            ResultClass=rc,
            Role=ro,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc))

        conn.OpenReferenceInstancePaths.assert_called_with(
            self.target_path(ns),
            ResultClass=rc,
            Role=ro,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)

        conn.CloseEnumeration.assert_not_called()

        # pylint: disable=protected-access
        assert conn._use_ref_path_pull_operations is True
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_paths

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    def test_pull_operation_success(self, tst_paths, use_pull_param):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterReferenceInstancePaths using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation. This must succeed.

        The Open operation returns no result objects, and the Pull operation
        returns all result objects and EOS.

        This does not retest the variations of input parameters because it
        is only confirming that the Pull operation executes.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        ctx = ('blah', conn.default_namespace)

        conn.ReferenceNames = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenReferenceInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=[], eos=False,
                                                     context=ctx))
        conn.PullInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths,
                                                     eos=True,
                                                     context=None))

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_ref_path_pull_operations == use_pull_param

        result = list(conn.IterReferenceInstancePaths(
            self.target_path('root/cimv2')))

        assert conn._use_ref_path_pull_operations is True
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_paths

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "max_obj_count", [1, 2]
    )
    @pytest.mark.parametrize(
        "close_after", [0, 1, 2]
    )
    def test_closed_operation_success(self, tst_paths, use_pull_param,
                                      max_obj_count, close_after):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterReferenceInstancePaths using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation.

        The iteration is closed after a number of retrievals, sometimes
        prematurely and sometimes not.

        This does not retest the variations of input parameters because it
        is only confirming that the Close operation executes if needed.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        tst_paths_open = tst_paths[0:max_obj_count]
        tst_paths_pull = tst_paths[max_obj_count:]
        ctx = ('blah', conn.default_namespace)

        conn.ReferenceNames = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenReferenceInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths_open,
                                                     eos=False,
                                                     context=ctx))
        conn.PullInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths_pull,
                                                     eos=True,
                                                     context=None))
        conn.CloseEnumeration = Mock()

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_ref_path_pull_operations == use_pull_param

        gen = conn.IterReferenceInstancePaths(
            self.target_path('root/cimv2'), MaxObjectCount=max_obj_count)
        result = []
        for _ in six.moves.range(close_after):
            obj = next(gen)
            result.append(obj)
        gen.close()  # close prematurely

        if close_after == 0 or close_after > max_obj_count:
            # Pull has been called; enumeration is exhausted; enumeration was
            # closed automatically within Pull.
            conn.CloseEnumeration.assert_not_called()
        else:
            # Pull has not been called; enumeration is not exhausted;
            # enumeration was closed by calling Close.
            conn.CloseEnumeration.assert_called_with(ctx)
            assert conn._use_ref_path_pull_operations is True

        assert conn.use_pull_operations == use_pull_param
        assert result == tst_paths[0:close_after]

    @pytest.mark.parametrize(
        "kwargs, exp_exc",
        [
            [dict(OperationTimeout='bla'), TypeError()],
            [dict(OperationTimeout=-1), ValueError()],
            [dict(MaxObjectCount=0), ValueError()],
            [dict(MaxObjectCount=-1), ValueError()],
            [dict(MaxObjectCount=None), ValueError()],
            [dict(MaxObjectCount='bla'), TypeError()],
            [dict(MaxObjectCount='bla'), TypeError()],
        ]
    )
    def test_open_invalid_params(self, tst_insts, kwargs, exp_exc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterReferenceInstancePaths using the pull operations,
        with invalid values and types of those input parameters that are
        actually checked (i.e. OperationTimeout and MaxObjectCount).
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.ReferenceNames = Mock(return_value=tst_insts)
        conn.OpenReferenceInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        with pytest.raises(type(exp_exc)):
            _ = list(conn.IterReferenceInstancePaths(
                self.target_path('dum'),
                **kwargs))

    @pytest.mark.parametrize(
        "kwargs, exp_exc",
        [
            [dict(FilterQueryLanguage='CQL'), ValueError()],
            [dict(FilterQuery='Prop=42'), ValueError()],
            [dict(ContinueOnError=True), ValueError()],
        ]
    )
    def test_trad_invalid_params(self, tst_insts, kwargs, exp_exc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterReferenceInstancePaths using the traditional operations,
        with invalid values and types of those input parameters that are
        actually checked (i.e. FilterQueryLanguage, FilterQuery,
        ContinueOnError).
        """
        conn = WBEMConnection('dummy', use_pull_operations=False)

        conn.ReferenceNames = Mock(return_value=tst_insts)
        conn.OpenReferenceInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        with pytest.raises(type(exp_exc)):
            _ = list(conn.IterReferenceInstancePaths(
                self.target_path('dum'),
                **kwargs))


########################################################################
#
#               IterAssociatorInstances tests
#
########################################################################

class TestIterAssociatorInstances(object):
    """Test IterAssociatorInstances execution"""

    @staticmethod
    def target_path(ns):
        """
        Return a valid target path for the calls
        """
        return CIMInstanceName('CIM_blah',
                               keybindings={'InstanceID': "1234"},
                               host='woot.com',
                               namespace=ns)

    @pytest.mark.parametrize(
        "use_pull_param", [None, False]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "iq, ico", [[None, None],
                    [False, False],
                    [True, True]]
    )
    # role, result_class, ResultRole, AssocClass
    @pytest.mark.parametrize(
        "ro, rc, rr, ac", [[None, None, None, None],
                           [None, 'CIM_Blah', None, 'AssocClass'],
                           ['RRole', None, 'ARole', None],
                           ['RRole', 'CIM_Blah', 'ARole', 'AssocClass']]
    )
    @pytest.mark.parametrize(
        "pl", [None, 'pl1', ['pl1', 'pl2']]
    )
    def test_trad_operation_success(self, use_pull_param, tst_insts, ns,
                                    rc, ro, ac, rr, iq, ico, pl,):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterAssociatorInstances using the traditional operations,
        either by requesting it via use_pull_operations=False, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the Open operation. This must succeed.

        The test is parameterized to test variations on all parameters. Note
        that it only tests legal parameters.

        The test confirms that the traditional operation receives the correct
        values for all parameters.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        # mock original function works, open returns CIMError
        conn.Associators = Mock(return_value=tst_insts)
        conn.OpenAssociatorInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_assoc_inst_pull_operations == use_pull_param

        result = list(conn.IterAssociatorInstances(
            self.target_path(ns),
            AssocClass=ac,
            ResultClass=rc,
            Role=ro,
            ResultRole=rr,
            IncludeQualifiers=iq,
            IncludeClassOrigin=ico,
            PropertyList=pl))

        conn.Associators.assert_called_with(
            self.target_path(ns),
            AssocClass=ac,
            ResultClass=rc,
            Role=ro,
            ResultRole=rr,
            IncludeQualifiers=iq,
            IncludeClassOrigin=ico,
            PropertyList=pl)

        assert conn._use_assoc_inst_pull_operations is False
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_insts

    def test_operation_fail(self, tst_insts):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterAssociatorInstances forcing the use of pull operations
        via use_pull_operations=True, and the Open operation raises
        NOT_SUPPORTED. This must fail.
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.Associators = Mock(return_value=tst_insts)
        conn.OpenAssociatorInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'description'))

        assert conn.use_pull_operations is True
        # pylint: disable=protected-access
        assert conn._use_assoc_inst_pull_operations is True

        with pytest.raises(CIMError) as exec_info:
            _ = list(conn.IterAssociatorInstances(
                self.target_path('root/cimv2')))

        exc = exec_info.value
        assert exc.status_code_name == 'CIM_ERR_NOT_SUPPORTED'

        # pylint: disable=protected-access
        assert conn._use_assoc_inst_pull_operations is True
        assert conn.use_pull_operations is True

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "iq, ico", [[None, None],
                    [False, False],
                    [True, True]]
    )
    # role, result_class, ResultRole, AssocClass
    @pytest.mark.parametrize(
        "ro, rc, rr, ac", [[None, None, None, None],
                           [None, 'CIM_Blah', None, 'AssocClass'],
                           ['RRole', None, 'ARole', None],
                           ['RRole', 'CIM_Blah', 'ARole', 'AssocClass']]
    )
    @pytest.mark.parametrize(
        "pl", [None, 'pl1', ['pl1', 'pl2']]
    )
    @pytest.mark.parametrize(
        "fl,fq", [(None, None), ('FQL', 'SELECT')]
    )
    @pytest.mark.parametrize(
        "ot, coe, moc", [[None, None, 100],
                         [10, False, 100]]
    )
    def test_open_operation_success(self, use_pull_param, tst_insts, ns,
                                    rc, ro, rr, ac, iq, ico, pl, fl, fq, ot,
                                    coe, moc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterAssociatorInstances using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation. This must succeed.

        The test is parameterized to test variations on all parameters. Note
        that it only tests legal parameters.

        The test confirms that the Open operation receives the correct
        values for all parameters, and that the Close operation is not called.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.Associators = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenAssociatorInstances = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts,
                                                     eos=True,
                                                     context=None))
        conn.CloseEnumeration = Mock()

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_assoc_inst_pull_operations == use_pull_param

        result = list(conn.IterAssociatorInstances(
            self.target_path(ns),
            AssocClass=ac,
            ResultClass=rc,
            Role=ro,
            ResultRole=rr,
            IncludeQualifiers=iq,
            IncludeClassOrigin=ico,
            PropertyList=pl,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc))

        conn.OpenAssociatorInstances.assert_called_with(
            self.target_path(ns),
            AssocClass=ac,
            ResultClass=rc,
            Role=ro,
            ResultRole=rr,
            IncludeClassOrigin=ico,
            PropertyList=pl,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)

        conn.CloseEnumeration.assert_not_called()

        # pylint: disable=protected-access
        assert conn._use_assoc_inst_pull_operations is True
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_insts

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    def test_pull_operation_success(self, tst_insts, use_pull_param):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterAssociatorInstances using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation. This must succeed.

        The Open operation returns no result objects, and the Pull operation
        returns all result objects and EOS.

        This does not retest the variations of input parameters because it
        is only confirming that the Pull operation executes.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        ctx = ('blah', conn.default_namespace)

        conn.Associators = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenAssociatorInstances = \
            Mock(return_value=pull_inst_result_tuple(instances=[], eos=False,
                                                     context=ctx))
        conn.PullInstancesWithPath = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts,
                                                     eos=True,
                                                     context=None))

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_assoc_inst_pull_operations == use_pull_param

        result = list(conn.IterAssociatorInstances(
            self.target_path('root/cimv2')))

        assert conn._use_assoc_inst_pull_operations is True
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_insts

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "max_obj_count", [1, 2]
    )
    @pytest.mark.parametrize(
        "close_after", [0, 1, 2]
    )
    def test_closed_operation_success(self, tst_insts, use_pull_param,
                                      max_obj_count, close_after):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterAssociatorInstances using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation.

        The iteration is closed after a number of retrievals, sometimes
        prematurely and sometimes not.

        This does not retest the variations of input parameters because it
        is only confirming that the Close operation executes if needed.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        tst_insts_open = tst_insts[0:max_obj_count]
        tst_insts_pull = tst_insts[max_obj_count:]
        ctx = ('blah', conn.default_namespace)

        conn.Associators = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenAssociatorInstances = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts_open,
                                                     eos=False,
                                                     context=ctx))
        conn.PullInstancesWithPath = \
            Mock(return_value=pull_inst_result_tuple(instances=tst_insts_pull,
                                                     eos=True,
                                                     context=None))
        conn.CloseEnumeration = Mock()

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_assoc_inst_pull_operations == use_pull_param

        gen = conn.IterAssociatorInstances(
            self.target_path('root/cimv2'), MaxObjectCount=max_obj_count)
        result = []
        for _ in six.moves.range(close_after):
            obj = next(gen)
            result.append(obj)
        gen.close()  # close prematurely

        if close_after == 0 or close_after > max_obj_count:
            # Pull has been called; enumeration is exhausted; enumeration was
            # closed automatically within Pull.
            conn.CloseEnumeration.assert_not_called()
        else:
            # Pull has not been called; enumeration is not exhausted;
            # enumeration was closed by calling Close.
            conn.CloseEnumeration.assert_called_with(ctx)
            assert conn._use_assoc_inst_pull_operations is True

        assert conn.use_pull_operations == use_pull_param
        assert result == tst_insts[0:close_after]

    @pytest.mark.parametrize(
        "kwargs, exp_exc",
        [
            [dict(OperationTimeout='bla'), TypeError()],
            [dict(OperationTimeout=-1), ValueError()],
            [dict(MaxObjectCount=0), ValueError()],
            [dict(MaxObjectCount=-1), ValueError()],
            [dict(MaxObjectCount=None), ValueError()],
            [dict(MaxObjectCount='bla'), TypeError()],
            [dict(MaxObjectCount='bla'), TypeError()],
        ]
    )
    def test_open_invalid_params(self, tst_insts, kwargs, exp_exc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterAssociatorInstances using the pull operations,
        with invalid values and types of those input parameters that are
        actually checked (i.e. OperationTimeout and MaxObjectCount).
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.Associators = Mock(return_value=tst_insts)
        conn.OpenAssociatorInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        with pytest.raises(type(exp_exc)):
            _ = list(conn.IterAssociatorInstances(
                self.target_path('root/x'),
                **kwargs))

    @pytest.mark.parametrize(
        "kwargs, exp_exc",
        [
            [dict(FilterQueryLanguage='CQL'), ValueError()],
            [dict(FilterQuery='Prop=42'), ValueError()],
            [dict(ContinueOnError=True), ValueError()],
        ]
    )
    def test_trad_invalid_params(self, tst_insts, kwargs, exp_exc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterAssociatorInstances using the traditional operations,
        with invalid values and types of those input parameters that are
        actually checked (i.e. FilterQueryLanguage, FilterQuery,
        ContinueOnError).
        """
        conn = WBEMConnection('dummy', use_pull_operations=False)

        conn.Associators = Mock(return_value=tst_insts)
        conn.OpenAssociatorInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        with pytest.raises(type(exp_exc)):
            _ = list(conn.IterAssociatorInstances(
                self.target_path('root/x'),
                **kwargs))


########################################################################
#
#               IterAssociatorInstancePaths tests
#
########################################################################

class TestIterAssociatorInstancePaths(object):
    """Test IterAssociatorInstancePaths execution"""

    @staticmethod
    def target_path(ns):
        """
            Return a valid target path for the calls
        """
        return CIMInstanceName('CIM_blah',
                               keybindings={'InstanceID': "1234"},
                               host='woot.com',
                               namespace=ns)

    @pytest.mark.parametrize(
        "use_pull_param", [None, False]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    # role, result_class, ResultRole, AssocClass
    @pytest.mark.parametrize(
        "ro, rc, rr, ac", [[None, None, None, None],
                           [None, 'CIM_Blah', None, 'AssocClass'],
                           ['RRole', None, 'ARole', None],
                           ['RRole', 'CIM_Blah', 'ARole', 'AssocClass']]
    )
    def test_trad_operation_success(self, use_pull_param, tst_paths, ns, rc,
                                    ro, rr, ac):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterAssociatorInstancePaths using the traditional operations,
        either by requesting it via use_pull_operations=False, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the Open operation. This must succeed.

        The test is parameterized to test variations on all parameters. Note
        that it only tests legal parameters.

        The test confirms that the traditional operation receives the correct
        values for all parameters.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        # mock original function works, open returns CIMError
        conn.AssociatorNames = Mock(return_value=tst_paths)
        conn.OpenAssociatorInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_assoc_path_pull_operations == use_pull_param

        result = list(conn.IterAssociatorInstancePaths(
            self.target_path(ns),
            AssocClass=ac,
            ResultClass=rc,
            Role=ro,
            ResultRole=rr))

        conn.AssociatorNames.assert_called_with(
            self.target_path(ns),
            AssocClass=ac,
            ResultClass=rc,
            Role=ro,
            ResultRole=rr)

        assert conn._use_assoc_path_pull_operations is False
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_paths

    def test_operation_fail(self, tst_paths):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterAssociatorInstancePaths forcing the use of pull operations
        via use_pull_operations=True, and the Open operation raises
        NOT_SUPPORTED. This must fail.
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.AssociatorNames = Mock(return_value=tst_paths)
        conn.OpenAssociatorInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'description'))

        assert conn.use_pull_operations is True
        # pylint: disable=protected-access
        assert conn._use_assoc_path_pull_operations is True

        with pytest.raises(CIMError) as exec_info:
            _ = list(conn.IterAssociatorInstancePaths(
                self.target_path('root/cimv2')))

        exc = exec_info.value
        assert exc.status_code_name == 'CIM_ERR_NOT_SUPPORTED'

        # pylint: disable=protected-access
        assert conn._use_assoc_path_pull_operations is True
        assert conn.use_pull_operations is True

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    # role, result_class, ResultRole, AssocClass
    @pytest.mark.parametrize(
        "ro, rc, rr, ac", [[None, None, None, None],
                           [None, 'CIM_Blah', None, 'AssocClass'],
                           ['RRole', None, 'ARole', None],
                           ['RRole', 'CIM_Blah', 'ARole', 'AssocClass']]
    )
    @pytest.mark.parametrize(
        "fl,fq", [(None, None), ('FQL', 'SELECT')]
    )
    @pytest.mark.parametrize(
        "ot, coe, moc", [[None, None, 100],
                         [10, False, 100]]
    )
    def test_open_operation_success(self, use_pull_param, tst_paths, ns,
                                    rc, ro, rr, ac, fl, fq, ot, coe, moc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterAssociatorInstancePaths using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation. This must succeed.

        The test is parameterized to test variations on all parameters. Note
        that it only tests legal parameters.

        The test confirms that the Open operation receives the correct
        values for all parameters, and that the Close operation is not called.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        conn.AssociatorNames = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenAssociatorInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths,
                                                     eos=True,
                                                     context=None))
        conn.CloseEnumeration = Mock()

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_assoc_path_pull_operations == use_pull_param

        result = list(conn.IterAssociatorInstancePaths(
            self.target_path(ns),
            AssocClass=ac,
            ResultClass=rc,
            Role=ro,
            ResultRole=rr,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc))

        conn.OpenAssociatorInstancePaths.assert_called_with(
            self.target_path(ns),
            AssocClass=ac,
            ResultClass=rc,
            Role=ro,
            ResultRole=rr,
            FilterQueryLanguage=fl,
            FilterQuery=fq,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)

        conn.CloseEnumeration.assert_not_called()

        # pylint: disable=protected-access
        assert conn._use_assoc_path_pull_operations is True
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_paths

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    def test_pull_operation_success(self, tst_paths, use_pull_param):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterAssociatorInstancePaths using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation. This must succeed.

        The Open operation returns no result objects, and the Pull operation
        returns all result objects and EOS.

        This does not retest the variations of input parameters because it
        is only confirming that the Pull operation executes.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        ctx = ('blah', conn.default_namespace)

        conn.AssociatorNames = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenAssociatorInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=[], eos=False,
                                                     context=ctx))
        conn.PullInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths,
                                                     eos=True,
                                                     context=None))

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_assoc_path_pull_operations == use_pull_param

        result = list(conn.IterAssociatorInstancePaths(
            self.target_path('root/cimv2')))

        assert conn._use_assoc_path_pull_operations is True
        assert conn.use_pull_operations == use_pull_param
        assert result == tst_paths

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "max_obj_count", [1, 2]
    )
    @pytest.mark.parametrize(
        "close_after", [0, 1, 2]
    )
    def test_closed_operation_success(self, tst_paths, use_pull_param,
                                      max_obj_count, close_after):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterAssociatorInstancePaths using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation.

        The iteration is closed after a number of retrievals, sometimes
        prematurely and sometimes not.

        This does not retest the variations of input parameters because it
        is only confirming that the Close operation executes if needed.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        tst_paths_open = tst_paths[0:max_obj_count]
        tst_paths_pull = tst_paths[max_obj_count:]
        ctx = ('blah', conn.default_namespace)

        conn.AssociatorNames = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenAssociatorInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths_open,
                                                     eos=False,
                                                     context=ctx))
        conn.PullInstancePaths = \
            Mock(return_value=pull_path_result_tuple(paths=tst_paths_pull,
                                                     eos=True,
                                                     context=None))
        conn.CloseEnumeration = Mock()

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_assoc_path_pull_operations == use_pull_param

        gen = conn.IterAssociatorInstancePaths(
            self.target_path('root/cimv2'), MaxObjectCount=max_obj_count)
        result = []
        for _ in six.moves.range(close_after):
            obj = next(gen)
            result.append(obj)
        gen.close()  # close prematurely

        if close_after == 0 or close_after > max_obj_count:
            # Pull has been called; enumeration is exhausted; enumeration was
            # closed automatically within Pull.
            conn.CloseEnumeration.assert_not_called()
        else:
            # Pull has not been called; enumeration is not exhausted;
            # enumeration was closed by calling Close.
            conn.CloseEnumeration.assert_called_with(ctx)
            assert conn._use_assoc_path_pull_operations is True

        assert conn.use_pull_operations == use_pull_param
        assert result == tst_paths[0:close_after]

    @pytest.mark.parametrize(
        "kwargs, exp_exc",
        [
            [dict(OperationTimeout='bla'), TypeError()],
            [dict(OperationTimeout=-1), ValueError()],
            [dict(MaxObjectCount=0), ValueError()],
            [dict(MaxObjectCount=-1), ValueError()],
            [dict(MaxObjectCount=None), ValueError()],
            [dict(MaxObjectCount='bla'), TypeError()],
            [dict(MaxObjectCount='bla'), TypeError()],
        ]
    )
    def test_open_invalid_params(self, tst_paths, kwargs, exp_exc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterAssociatorInstancePaths using the pull operations,
        with invalid values and types of those input parameters that are
        actually checked (i.e. OperationTimeout and MaxObjectCount).
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.AssociatorNames = Mock(return_value=tst_paths)
        conn.OpenAssociatorInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        with pytest.raises(type(exp_exc)):
            _ = list(conn.IterAssociatorInstancePaths(
                self.target_path('dum'),
                **kwargs))

    @pytest.mark.parametrize(
        "kwargs, exp_exc",
        [
            [dict(FilterQueryLanguage='CQL'), ValueError()],
            [dict(FilterQuery='Prop=42'), ValueError()],
            [dict(ContinueOnError=True), ValueError()],
        ]
    )
    def test_trad_invalid_params(self, tst_paths, kwargs, exp_exc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterAssociatorInstancePaths using the traditional operations,
        with invalid values and types of those input parameters that are
        actually checked (i.e. FilterQueryLanguage, FilterQuery,
        ContinueOnError).
        """
        conn = WBEMConnection('dummy', use_pull_operations=False)

        conn.AssociatorNames = Mock(return_value=tst_paths)
        conn.OpenAssociatorInstancePaths = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        with pytest.raises(type(exp_exc)):
            _ = list(conn.IterAssociatorInstancePaths(
                self.target_path('dum'),
                **kwargs))


########################################################################
#
#               IterQueryInstances tests
#
########################################################################

class TestIterQueryInstances(object):
    """Test IterQueryInstances execution"""

    @pytest.mark.parametrize(
        "use_pull_param", [None, False]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "ql, query", [(None, None), ('CQL', 'SELECT from *')]
    )
    def test_trad_operation_success(self, use_pull_param, tst_insts, ns,
                                    ql, query):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterQueryInstances using the traditional operations,
        either by requesting it via use_pull_operations=False, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the Open operation. This must succeed.

        The test is parameterized to test variations on all parameters. Note
        that it only tests legal parameters.

        The test confirms that the traditional operation receives the correct
        values for all parameters.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        # mock original function works, open returns CIMError
        conn.ExecQuery = Mock(return_value=tst_insts)
        conn.OpenQueryInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_query_pull_operations == use_pull_param

        q_result = conn.IterQueryInstances(ql, query, namespace=ns)
        result_insts = list(q_result.generator)

        conn.ExecQuery.assert_called_with(ql, query, namespace=ns)

        assert q_result.query_result_class is None
        assert conn._use_query_pull_operations is False
        assert conn.use_pull_operations == use_pull_param
        assert result_insts == tst_insts

    def test_operation_fail(self, tst_insts):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterQueryInstances forcing the use of pull operations
        via use_pull_operations=True, and the Open operation raises
        NOT_SUPPORTED. This must fail.
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.ExecQuery = Mock(return_value=tst_insts)
        conn.OpenQueryInstances = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'description'))

        assert conn.use_pull_operations is True
        # pylint: disable=protected-access
        assert conn._use_query_pull_operations is True

        with pytest.raises(CIMError) as exec_info:
            conn.IterQueryInstances('CQL', 'Select from *')

        exc = exec_info.value
        assert exc.status_code_name == 'CIM_ERR_NOT_SUPPORTED'

        # pylint: disable=protected-access
        assert conn._use_query_pull_operations is True
        assert conn.use_pull_operations is True

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "ns", [None, 'test/testnamespace']
    )
    @pytest.mark.parametrize(
        "ql, query", [(None, None), ('CQL', 'SELECT from *')]
    )
    @pytest.mark.parametrize(
        "ot, coe, moc", [[None, None, 100],
                         [10, False, 100]]
    )
    # ReturnQueryResultClass (Boolean flag))
    @pytest.mark.parametrize(
        "rqrc_param", [None, True]
    )
    def test_open_operation_success(self, use_pull_param, tst_insts, ns,
                                    ql, query, ot, coe, moc, rqrc_param,
                                    tst_class):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterQueryInstances using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation. This must succeed.

        The test is parameterized to test variations on all parameters. Note
        that it only tests legal parameters.

        The test confirms that the Open operation receives the correct
        values for all parameters, and that the Close operation is not called.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        # Get class definition for response if ReturnQueryResultClass
        qrc = tst_class if rqrc_param is True else None

        conn.ExecQuery = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenQueryInstances = \
            Mock(return_value=pull_query_result_tuple(instances=tst_insts,
                                                      eos=True,
                                                      context=None,
                                                      query_result_class=qrc))
        conn.CloseEnumeration = Mock()

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_query_pull_operations == use_pull_param

        q_result = conn.IterQueryInstances(
            ql, query,
            namespace=ns,
            ReturnQueryResultClass=rqrc_param,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)

        conn.OpenQueryInstances.assert_called_with(
            ql, query,
            namespace=ns,
            ReturnQueryResultClass=rqrc_param,
            OperationTimeout=ot,
            ContinueOnError=coe,
            MaxObjectCount=moc)

        conn.CloseEnumeration.assert_not_called()

        result_insts = list(q_result.generator)

        assert q_result.query_result_class == qrc

        # pylint: disable=protected-access
        assert conn._use_query_pull_operations is True
        assert conn.use_pull_operations == use_pull_param
        assert result_insts == tst_insts

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    # ReturnQueryResultClass
    @pytest.mark.parametrize(
        "rqrc_param", [None, True]
    )
    def test_pull_operation_success(self, tst_insts, use_pull_param,
                                    rqrc_param):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterQueryInstances using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation. This must succeed.

        The Open operation returns no result objects, and the Pull operation
        returns all result objects and EOS.

        This does not retest the variations of input parameters because it
        is only confirming that the Pull operation executes.

        The test confirms that the Open operation receives the correct
        values for all parameters, and that the Close operation is not called.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        # Get class definition for response if ReturnQueryResultClass
        qrc = tst_class if rqrc_param is True else None

        conn.ExecQuery = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenQueryInstances = \
            Mock(return_value=pull_query_result_tuple(
                instances=[],
                eos=False,
                context="BlahBlah",
                query_result_class=qrc))
        conn.PullInstances = \
            Mock(return_value=pull_query_result_tuple(instances=tst_insts,
                                                      eos=True,
                                                      context=None,
                                                      query_result_class=qrc))
        conn.CloseEnumeration = Mock()

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_assoc_path_pull_operations == use_pull_param

        q_result = conn.IterQueryInstances(
            'CQL', 'Select from *',
            ReturnQueryResultClass=rqrc_param)

        result_insts = list(q_result.generator)

        conn.OpenQueryInstances.assert_called_with(
            'CQL', 'Select from *',
            namespace=None,
            ReturnQueryResultClass=rqrc_param,
            OperationTimeout=None,
            ContinueOnError=None,
            MaxObjectCount=DEFAULT_ITER_MAXOBJECTCOUNT)

        conn.CloseEnumeration.assert_not_called()

        # TODO ks: This assert disabled.
        # assert(q_result.query_result_class == rc)

        # pylint: disable=protected-access
        assert conn._use_query_pull_operations is True
        assert conn.use_pull_operations == use_pull_param
        assert result_insts == tst_insts

    @pytest.mark.parametrize(
        "use_pull_param", [None, True]
    )
    @pytest.mark.parametrize(
        "max_obj_count", [1, 2]
    )
    @pytest.mark.parametrize(
        "close_after", [0, 1, 2]
    )
    def test_closed_operation_success(self, tst_insts, use_pull_param,
                                      max_obj_count, close_after):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterQueryInstances using the pull operations,
        either by requesting it via use_pull_operations=True, or by leaving it
        to the implementation via use_pull_operations=None and raising
        NOT_SUPPORTED in the traditional operation.

        The iteration is closed after a number of retrievals, sometimes
        prematurely and sometimes not.

        This does not retest the variations of input parameters because it
        is only confirming that the Close operation executes if needed.
        """
        conn = WBEMConnection('dummy', use_pull_operations=use_pull_param)

        tst_insts_open = tst_insts[0:max_obj_count]
        tst_insts_pull = tst_insts[max_obj_count:]
        ctx = ('blah', conn.default_namespace)

        conn.ExecQuery = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenQueryInstances = \
            Mock(return_value=pull_query_result_tuple(instances=tst_insts_open,
                                                      eos=False,
                                                      context=ctx,
                                                      query_result_class=None))
        conn.PullInstances = \
            Mock(return_value=pull_query_result_tuple(instances=tst_insts_pull,
                                                      eos=True,
                                                      context=None,
                                                      query_result_class=None))
        conn.CloseEnumeration = Mock()

        assert conn.use_pull_operations == use_pull_param
        # pylint: disable=protected-access
        assert conn._use_query_pull_operations == use_pull_param

        q_result = conn.IterQueryInstances(
            'CQL', 'Select from *',
            ReturnQueryResultClass=None, MaxObjectCount=max_obj_count)
        gen = q_result.generator
        result = []
        for _ in six.moves.range(close_after):
            obj = next(gen)
            result.append(obj)
        gen.close()  # close prematurely

        # In the current implementation of IterQuery, the enumeration is
        # always exhausted.
        conn.CloseEnumeration.assert_not_called()

        assert conn.use_pull_operations == use_pull_param
        assert result == tst_insts[0:close_after]

    @pytest.mark.parametrize(
        "kwargs, exp_exc",
        [
            [dict(OperationTimeout='bla'), TypeError()],
            [dict(OperationTimeout=-1), ValueError()],
            [dict(MaxObjectCount=0), ValueError()],
            [dict(MaxObjectCount=-1), ValueError()],
            [dict(MaxObjectCount=None), ValueError()],
            [dict(MaxObjectCount='bla'), TypeError()],
            [dict(MaxObjectCount='bla'), TypeError()],
        ]
    )
    def test_open_invalid_params(self, tst_insts, kwargs, exp_exc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterQueryInstances using the pull operations,
        with invalid values and types of those input parameters that are
        actually checked (i.e. OperationTimeout and MaxObjectCount).
        """
        conn = WBEMConnection('dummy', use_pull_operations=True)

        conn.ExecQuery = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenQueryInstances = \
            Mock(return_value=pull_query_result_tuple(instances=tst_insts,
                                                      eos=True,
                                                      context=None,
                                                      query_result_class=None))
        with pytest.raises(type(exp_exc)):
            conn.IterQueryInstances(
                'CQL', 'Select from *',
                **kwargs)

    @pytest.mark.parametrize(
        "kwargs, exp_exc",
        [
            [dict(ContinueOnError=True), ValueError()],
            [dict(ReturnQueryResultClass=True), ValueError()],
        ]
    )
    def test_trad_invalid_params(self, tst_insts, kwargs, exp_exc):
        # pylint: disable=no-self-use,redefined-outer-name
        """
        Test IterQueryInstances using the traditional operations,
        with invalid values and types of those input parameters that are
        actually checked (i.e. ContinueOnError).
        """
        conn = WBEMConnection('dummy', use_pull_operations=False)

        conn.ExecQuery = \
            Mock(side_effect=CIMError(CIM_ERR_NOT_SUPPORTED, 'Blah'))
        conn.OpenQueryInstances = \
            Mock(return_value=pull_query_result_tuple(instances=tst_insts,
                                                      eos=True,
                                                      context=None,
                                                      query_result_class=None))
        with pytest.raises(type(exp_exc)):
            conn.IterQueryInstances(
                'CQL', 'Select from *',
                **kwargs)



#
# (C) Copyright 2017 InovaDevelopment Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# pylint: disable=line-too-long
"""
*New in pywbem 0.11 as experimental and finalized in 0.12.*

Pywbem supports measuring the elapsed times of the WBEM operations that
were performed in context of a connection, and maintaining a statistics
over these times.

This capability is disabled by default and can be enabled in either
of these ways:

* When creating a :class:`~pywbem.WBEMConnection` object, via its
  ``stats_enabled`` argument.

* After the :class:`~pywbem.WBEMConnection` object has been created, by
  modifying its :attr:`~pywbem.WBEMConnection.stats_enabled` instance
  attribute.

The :class:`~pywbem.Statistics` class maintains statistics over the measured
elapsed times of the WBEM operations and is the interface for accessing the
statistics. The statistics of a :class:`~pywbem.WBEMConnection` object are
accessible via its :attr:`~pywbem.WBEMConnection.statistics` instance
attribute.

The :class:`~pywbem.OperationStatistic` class is a helper class that contains
the actual measurement data for one operation name.
There will be one :class:`~pywbem.OperationStatistic` object each operation
name (see the table of WBEMConnection methods in the :ref:`WBEM operations`
section for the operation names).
The :class:`~pywbem.OperationStatistic` objects are under control of the
:class:`~pywbem.Statistics` class.

The statistics support maintains two kinds of times for each kind of WBEM
operation:

* Client times: The elapsed times for the WBEMConnection operation methods from
  call to return. This is measured in pywbem pretty close to the API the pywbem
  user is calling.

* Server times: The elapsed times for processing the WBEM operations in the
  WBEM server from receiving the CIM-XML request message to sending back the
  CIM-XML response message. The server times are not measured by pywbem, but
  are taken from the `WBEMServerResponseTime` HTTP header field of a CIM-XML
  response, if present. See :term:`DSP0200` for a description of this header
  field.

  The `WBEMServerResponseTime` HTTP header field is optionally implemented
  by WBEM servers. The following WBEM servers are known to implement this
  header field:

  * OpenPegasus

The difference between client time and server time is the time spent in the
pywbem client, plus the time spent on the network between client and server.

The statistics support also maintains the size of the HTTP body in the CIM-XML
request and response messages, in Bytes.

These times and sizes are maintained as average, minimum and maximum values for
each kind of operation in a connection.

Finally, the statistics support maintains the total count of operations and the
count of operations that failed, for each kind of operation.

All data in the statistics applies to WBEM operations performed during periods
of time where the statistics are enabled on a connection. Operations performed
during periods of time where the statistics are disabled on a connection, are
simply ignored in the statistics.

For the Iter methods of WBEMConnection (e.g.
:meth:`~pywbem.WbemConnection.IterEnumerateInstances`), the WBEM operations
performed on behalf of them are subject of the statistics, but the Iter methods
themselves do not show up in the statistics.

The following example shows how statistics are enabled, and how statistics
values are accessed individually using the
:meth:`~pywbem.Statistics.get_op_statistic` method::

    conn = pywbem.WBEMConnection(..., stats_enabled=True)

    # Perform some operations on this connection
    insts_1 = conn.EnumerateInstances('CIM_Foo1', 'root/cimv2')
    insts_2 = conn.EnumerateInstances('CIM_Foo2', 'root/cimv2')
    insts_3 = conn.EnumerateInstances('CIM_Foo3', 'root/cimv2')
    inst_paths_1 = conn.EnumerateInstanceNames('CIM_Foo1', 'root/cimv2')

    # Access statistics values for EnumerateInstances
    ei_stats = conn.statistics.get_op_statistic('EnumerateInstances')
    ei_count = ei_stats.count
    ei_avg_client_time = ei_stats.avg_time
    ei_avg_server_time = ei_stats.avg_server_time

In the previous example, the values in ``ei_stats`` are "live", i.e. they
continue to be updated as operations are performed. If a snapshot is needed at
a certain point in time that remains unaffected by further operations, this can
be achieved using the :meth:`~pywbem.Statistics.snapshot` method::

    # Take snapshot and access statistics values for EnumerateInstances
    stats_snapshot = dict(conn.statistics.snapshot())
    ei_stats = stats_snapshot['EnumerateInstances']
    ei_count = ei_stats.count
    ei_avg_client_time = ei_stats.avg_time
    ei_avg_server_time = ei_stats.avg_server_time

It is also possible to simply print the current statistics of a connection as a
formatted table, using the :meth:`~pywbem.Statistics.formatted` method::

    # Print statistics values for all operations
    print(conn.statistics.formatted())

The output could look like this, if the WBEM server returns WBEM server
response times::

    Statistics (times in seconds, lengths in Bytes):
    Count Excep         ClientTime              ServerTime             RequestLen                ReplyLen       Operation
            Cnt     Avg     Min     Max     Avg     Min     Max    Avg    Min    Max      Avg      Min      Max
        3     0   0.234   0.100   0.401   0.204   0.080   0.361   1233   1000   1500    26667    20000    35000 EnumerateInstances
        1     0   0.100   0.100   0.100   0.080   0.080   0.080   1200   1200   1200    22000    22000    22000 EnumerateInstanceNames
"""  # noqa: E501
# pylint: enable=line-too-long

from __future__ import absolute_import

import time
import copy
import six

from ._utils import _format

__all__ = ['Statistics', 'OperationStatistic']


class OperationStatistic(object):
    # pylint: disable=too-many-instance-attributes
    """
    *New in pywbem 0.11 as experimental and finalized in 0.12.*

    A statistics data keeper for the executions of all operations with the
    same operation name.

    Objects of this class are created by the :class:`~pywbem.Statistics` class
    and can be accessed by pywbem users through its
    :meth:`~pywbem.Statistics.get_op_statistic` and
    :meth:`~pywbem.Statistics.snapshot` methods.
    """

    def __init__(self, container, name):
        """
        Parameters:

          container (:class:`~pywbem.Statistics`):
            The statistics container that holds this operation statistic
            object.

          name (string):
            Name of the operation.
        """
        self._container = container
        self._stat_start_time = None
        self._name = name
        self._count = 0
        self._exception_count = 0
        self._time_sum = float(0)
        self._time_min = float('inf')
        self._time_max = float(0)

        self._server_time_sum = float(0)
        self._server_time_min = float('inf')
        self._server_time_max = float(0)
        self._server_time_stored = False

        self._start_time = None

        self._request_len_sum = float(0)
        self._request_len_min = float('inf')
        self._request_len_max = float(0)

        self._reply_len_sum = float(0)
        self._reply_len_min = float('inf')
        self._reply_len_max = float(0)

    @property
    def stat_start_time(self):
        """
        :class:`py:float`: Point in time when the first statistic was taken
        since this object was either created or reset, in seconds since the
        epoch (for details, see :func:`py:time.time`).
        """
        return self._stat_start_time

    @property
    def name(self):
        """
        :term:`string`: Name of the operation for which this statistics
        object maintains data.
        """
        return self._name

    @property
    def container(self):
        """
        :class:`~pywbem.Statistics`: The statistics container that holds
        this statistics object.
        """
        return self._container

    @property
    def count(self):
        """
        :term:`integer`: The number of measured operations.
        """
        return self._count

    @property
    def exception_count(self):
        """
        :term:`integer`: The number of measured operations that resulted in an
        exception returned to the pywbem user (for example because a failure
        was indicated in the operation response of the WBEM server, or because
        pywbem itself detected a failure before sending the request or after
        receiving the response).
        """
        return self._exception_count

    @property
    def avg_time(self):
        """
        :class:`py:float`: The average elapsed client time for execution of the
        measured operations, in seconds.
        """
        try:
            return self._time_sum / self._count
        except ZeroDivisionError:
            return 0

    @property
    def min_time(self):
        """
        :class:`py:float`: The minimum elapsed client time for execution of the
        measured operations, in seconds.
        """
        return self._time_min

    @property
    def max_time(self):
        """
        :class:`py:float`: The maximum elapsed client time for execution of the
        measured operations, in seconds.
        """
        return self._time_max

    @property
    def avg_server_time(self):
        """
        :class:`py:float`: The average elapsed server time for execution of the
        measured operations, in seconds.

        This time is 0 if the WBEM server did not return the WBEM server
        response time.
        """
        try:
            return self._server_time_sum / self._count
        except ZeroDivisionError:
            return 0

    @property
    def min_server_time(self):
        """
        :class:`py:float`: The minimum elapsed server time for execution of the
        measured operations, in seconds.

        This time is 0 if the WBEM server did not return the WBEM server
        response time.
        """
        return self._server_time_min

    @property
    def max_server_time(self):
        """
        :class:`py:float`: The maximum elapsed server time for execution of the
        measured operations, in seconds.

        This time is 0 if the WBEM server did not return the WBEM server
        response time.
        """
        return self._server_time_max

    @property
    def avg_request_len(self):
        """
        :class:`py:float`: The average size of the HTTP body in the CIM-XML
        requests of the measured operations, in Bytes.
        """
        try:
            return self._request_len_sum / self._count
        except ZeroDivisionError:
            return 0.0

    @property
    def min_request_len(self):
        """
        :class:`py:float`: The minimum size of the HTTP body in the CIM-XML
        requests of the measured operations, in Bytes.
        """
        return self._request_len_min

    @property
    def max_request_len(self):
        """
        :class:`py:float`: The maximum size of the HTTP body in the CIM-XML
        requests of the measured operations, in Bytes.
        """
        return self._request_len_max

    @property
    def avg_reply_len(self):
        """
        :class:`py:float`: The average size of the HTTP body in the CIM-XML
        responses of the measured operations, in Bytes.
        """
        try:
            return self._reply_len_sum / self._count
        except ZeroDivisionError:
            return 0.0

    @property
    def min_reply_len(self):
        """
        :class:`py:float`: The minimum size of the HTTP body in the CIM-XML
        responses of the measured operations, in Bytes.
        """
        return self._reply_len_min

    @property
    def max_reply_len(self):
        """
        :class:`py:float`: The maximum size of the HTTP body in the CIM-XML
        responses of the measured operations, in Bytes.
        """
        return self._reply_len_max

    def reset(self):
        """
        Reset the statistics data for this object.
        """
        self._count = 0
        self._exception_count = 0
        self._stat_start_time = None
        self._time_sum = float(0)
        self._time_min = float('inf')
        self._time_max = float(0)

        self._server_time_sum = float(0)
        self._server_time_min = float('inf')
        self._server_time_max = float(0)
        self._server_time_stored = False

        self._request_len_sum = float(0)
        self._request_len_min = float('inf')
        self._request_len_max = float(0)

        self._reply_len_sum = float(0)
        self._reply_len_min = float('inf')
        self._reply_len_max = float(0)

    def start_timer(self):
        """
        This is a low-level method that is called by pywbem at the begin of an
        operation. It starts the measurement for that operation, if statistics
        is enabled for the connection.

        A subsequent invocation of :meth:`~pywbem.OperationStatistic.stop_timer`
        will complete the measurement for that operation and will update the
        statistics data.
        """
        if self.container.enabled:
            self._start_time = time.time()
            if not self._stat_start_time:
                self._stat_start_time = self._start_time

    def stop_timer(self, request_len=None, reply_len=None, server_time=None,
                   exception=False):
        """
        This is a low-level method is called by pywbem at the end of an
        operation. It completes the measurement for that operation by capturing
        the needed data, and updates the statistics data, if statistics is
        enabled for the connection.

        Parameters:

          request_len (:term:`integer`):
            Size of the HTTP body of the CIM-XML request message, in Bytes.

          reply_len (:term:`integer`):
            Size of the HTTP body of the CIM-XML response message, in Bytes.

          exception (:class:`py:bool`):
            Boolean that specifies whether an exception was raised while
            processing the operation.

          server_time (:class:`py:bool`):
            Time in seconds that the server optionally returns to the
            client in the HTTP response defining the time from when the
            server received the request to when it started sending the
            response. If `None`, there is no time from the server.

        Returns:

          float: The elapsed time for the operation that just ended, or
          `None` if the statistics container holding this object is not
          enabled.
        """
        if not self.container.enabled:
            return None

        # stop the timer
        if self._start_time is None:
            raise RuntimeError('stop_timer() called without preceding '
                               'start_timer()')
        dt = time.time() - self._start_time
        self._start_time = None
        self._count += 1
        self._time_sum += dt
        if request_len is not None:
            self._request_len_sum += request_len
        if reply_len is not None:
            self._reply_len_sum += reply_len

        if exception:
            self._exception_count += 1

        if dt > self._time_max:
            self._time_max = dt
        if dt < self._time_min:
            self._time_min = dt

        if server_time:
            self._server_time_stored = True
            self._server_time_sum += server_time
            if dt > self._server_time_max:
                self._server_time_max = server_time
            if dt < self._server_time_min:
                self._server_time_min = server_time

        if request_len is not None:
            if request_len > self._request_len_max:
                self._request_len_max = request_len
            if request_len < self._request_len_min:
                self._request_len_min = request_len

        if reply_len is not None:
            if reply_len > self._reply_len_max:
                self._reply_len_max = reply_len
            if reply_len < self._reply_len_min:
                self._reply_len_min = reply_len

        return dt

    def __repr__(self):
        """
        Return a human readable string with the statistics values, for debug
        purposes.
        """
        return _format(
            "OperationStatistic("
            "name={s.name!A}, "
            "count={s.count!A}, "
            "exception_count={s.exception_count!A}, "
            "avg_time={s.avg_time!A}, "
            "min_time={s.min_time!A}, "
            "max_time={s.max_time!A}, "
            "avg_server_time={s.avg_server_time!A}, "
            "min_server_time={s.min_server_time!A}, "
            "max_server_time={s.max_server_time!A}, "
            "avg_request_len={s.avg_request_len!A}, "
            "min_request_len={s.min_request_len!A}, "
            "max_request_len={s.max_request_len!A}, "
            "avg_reply_len={s.avg_reply_len!A}, "
            "min_reply_len={s.min_reply_len!A}, "
            "max_reply_len={s.max_reply_len!A})",
            s=self)

    @staticmethod
    def formatted_header(include_server_time, include_lengths):
        """
        Return a two-line header.
        """
        ret_lines = [
            'Count ExcCnt          Time [s]      ',
            '                Avg     Min     Max ',
        ]
        if include_server_time:
            ret_lines[0] += '       ServerTime [s]   '
            ret_lines[1] += '    Avg     Min     Max '
        if include_lengths:
            ret_lines[0] += '       RequestLen [B]            ReplyLen [B]   '
            ret_lines[1] += '   Avg    Min    Max      Avg      Min      Max '
        ret_lines[0] += 'Operation\n'
        ret_lines[1] += '\n'
        return ''.join(ret_lines)

    def formatted(self, include_server_time, include_lengths):
        """
        Return a formatted one-line string with the statistics
        values for the operation for which this statistics object
        maintains data.

        This is a low-level method that is called by
        :meth:`pywbem.Statistics.formatted`.
        """
        ret = ('{0:5d} {1:5d} {2:7.3f} {3:7.3f} {4:7.3f} '.
               format(self.count, self.exception_count,
                      self.avg_time, self.min_time, self.max_time))
        if include_server_time:
            ret += ('{0:7.3f} {1:7.3f} {2:7.3f} '.
                    format(self.avg_server_time, self.min_server_time,
                           self.max_server_time))
        if include_lengths:
            ret += ('{0:6.0f} {1:6.0f} {2:6.0f} '
                    '{3:8.0f} {4:8.0f} {5:8.0f} '.
                    format(self.avg_request_len, self.min_request_len,
                           self.max_request_len, self.avg_reply_len,
                           self.min_reply_len, self.max_reply_len))
        ret += '{0}\n'.format(self.name)
        return ret


class Statistics(object):
    """
    *New in pywbem 0.11 as experimental and finalized in 0.12.*

    The statistics of a connection, that captures and provides statistics data
    about the WBEM operations performed on the connection.

    This class contains an operation statistic object (of class
    :class:`~pywbem.OperationStatistic`) for each kind of WBEM operation.

    A :class:`~pywbem.Statistics` object can be in a state of enabled or
    disabled.
    If enabled, it accumulates the elapsed times between subsequent calls to
    the :meth:`~pywbem.OperationStatistic.start_timer` and
    :meth:`~pywbem.OperationStatistic.stop_timer` methods of class
    :class:`~pywbem.OperationStatistic`.
    If disabled, calls to these methods do not accumulate any time.
    Initially, the statistics container is disabled.

    The enablement state of the :class:`~pywbem.Statistics` object is
    controlled by the statistics enablement state of the connection it belongs
    to (see :meth:`pywbem.WBEMConnection.stats_enabled`)

    This class can also be used as a context manager.
    """

    def __init__(self, enable=False):
        """
        Parameters:

          enable (:class:`py:bool`):
            Initial enablement status for this statistics container.
        """
        # We convert any non-boolean values to True/False:
        self._enabled = bool(enable)
        self._op_stats = {}
        self._disabled_stats = OperationStatistic(self, "disabled")

        # Used in context manager (which supports nesting)
        self._cm_stack = []  # items: OperationStatistic object
        self._cm_name = None  # stored only between __call__() and __enter__()

    def __enter__(self):
        """
        Enter method when the class is used as a context manager.

        Starts the operation statistics for the name that was previously set
        via a call::

            stats = Statistics()
            with stats(name='bla'):
                # do something

        The class supports nesting of context managers::

            stats = Statistics()
            with stats(name='bla1'):
                # do something
                for i in ...:
                    with stats(name='bla2'):
                        # do something
        """
        name = self._cm_name
        self._cm_name = None
        op_stat = self.start_timer(name)
        self._cm_stack.append(op_stat)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit method when the class is used as a context manager.

        Stops the operation statistics that was started in the enter method.
        """
        op_stat = self._cm_stack.pop()
        op_stat.stop_timer()
        return False  # re-raise any exceptions

    def __call__(self, name):
        """
        This allows the `name` parameter to be passed when the class is used
        as a context manager.
        """
        self._cm_name = name
        return self

    @property
    def enabled(self):
        """
        Indicates whether the statistics container is enabled.
        """
        return self._enabled

    def enable(self):
        """
        Enable the statistics container.
        """
        self._enabled = True

    def disable(self):
        """
        Disable the statistics container.
        """
        self._enabled = False

    def start_timer(self, name):
        """
        This method is called by pywbem to start the timer for a particular
        operation execution. It returns the corresponding
        :class:`~pywbem.OperationStatistic` object, creating one if needed.

        The timer is subsequently stopped by pywbem by calling the
        :meth:`~pywbem.OperationStatistic.stop_timer` method on the returned
        :class:`~pywbem.OperationStatistic` object.

        Parameters:

          name (string):
            Name of the operation.

        Returns:

          :class:`~pywbem.OperationStatistic`: The operation statistic for the
          specified name. If this statistics container is disabled, a dummy
          operation statistic object is returned.
        """
        op_stat = self.get_op_statistic(name)
        op_stat.start_timer()
        return op_stat

    def get_op_statistic(self, name):
        """
        Get the :class:`~pywbem.OperationStatistic` object for an operation
        name or create a new object if an object for that name does not exist.

        Parameters:

          name (string):
            Name of the operation.

        Returns:

          :class:`~pywbem.OperationStatistic`: The operation statistic for the
          specified operation name. If this statistics container is disabled,
          a dummy operation statistic object is returned.
        """
        if not self.enabled:
            return self._disabled_stats
        if name not in self._op_stats:
            self._op_stats[name] = OperationStatistic(self, name)
        return self._op_stats[name]

    def snapshot(self):
        """
        Return a snapshot of the time statistics of this container.

        The snapshot represents the statistics data at the time this method
        is called, and remains unchanged even if the statistics of this
        container continues to be updated.

        Returns:

          : A list of tuples (name, stats) with:

          - name (:term:`string`): Name of the operation
          - stats (:class:`~pywbem.OperationStatistic`): Time statistics for
            the operation
        """
        return list(copy.deepcopy(self._op_stats).items())

    def __repr__(self):
        """
        Return a human readable display of the contents, for debug purposes.
        """
        ret = "Statistics(\n"
        for name in self._op_stats:
            ret += _format("  {0!A}\n", self._op_stats[name])
        ret += ")"
        return ret

    def formatted(self):
        # pylint: disable=line-too-long
        """
        Return a human readable string with the statistics for this container.
        The operations are sorted by decreasing average time.

        The three columns for `ServerTime` are included only if the WBEM server
        has returned WBEM server response times.

        The six columns for `RequestLen` and `ReplyLen` are included only if
        they are non-zero (this allows using this class for other purposes).

        Example if statistics are enabled::

            Statistics:
            Count ExcCnt         Time [s]             ServerTime [s]        RequestLen [B]            ReplyLen [B]      Operation
                            Avg     Min     Max     Avg     Min     Max    Avg    Min    Max      Avg      Min      Max
                3     0   0.234   0.100   0.401   0.204   0.080   0.361   1233   1000   1500    26667    20000    35000 EnumerateInstances
                1     0   0.100   0.100   0.100   0.080   0.080   0.080   1200   1200   1200    22000    22000    22000 EnumerateInstanceNames
                . . .

        Example if statistics are disabled::

            Statistics:
            Disabled
        """  # noqa: E501
        # pylint: enable=line-too-long
        ret = "Statistics:\n"
        if self.enabled:
            snapshot = sorted(self.snapshot(),
                              key=lambda item: item[1].avg_time,
                              reverse=True)

            # Test to see if any server time is non-zero
            include_svr = False
            include_len = False
            for _, stats in snapshot:
                # pylint: disable=protected-access
                if stats._server_time_stored:
                    include_svr = True
                # pylint: disable=protected-access
                if stats._request_len_sum > 0 or stats._reply_len_sum > 0:
                    include_len = True

            ret += OperationStatistic.formatted_header(
                include_svr, include_len)
            for _, stats in snapshot:
                ret += stats.formatted(include_svr, include_len)

        else:
            ret += "Disabled"
        return ret.strip()

    def reset(self):
        """
        Reset all statistics and clear any statistic names.
        All statistics must be inactive before a reset will execute

        Returns: True if reset, False if not
        """
        # Test for any stats being currently timed.
        for stat in six.itervalues(self._op_stats):
            if stat._start_time is not None:  # pylint: disable=protected-access
                return False

        # clear all statistics
        self._op_stats = {}
        return True



#
# (C) Copyright 2020 InovaDevelopment.comn
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Karl  Schopmeyer <inovadevelopment.com>
#

"""
The class :class:`~pywbem_mock.InMemoryRepository` implements a CIM repository
that stores the CIM objects in memory. It contains an object store
(:class:`~pywbem_mock.InMemoryObjectStore`) for each type of CIM object
(classes, instances, qualifier declarations) and for each CIM namespace in the
repository. Each object store for a CIM object type contains the CIM objects
of that type in a single CIM namespace.

The CIM repository of a mock WBEM server represented by a
:class:`~pywbem_mock.FakedWBEMConnection` object is a
:class:`~pywbem_mock.InMemoryRepository` object that is created and destroyed
automatically when the
:class:`~pywbem_mock.FakedWBEMConnection` object is created and destroyed.

The CIM repository of a mock WBEM server is accessible through its
:attr:`pywbem_mock.FakedWBEMConnection.cimrepository` property.
"""

from __future__ import absolute_import, print_function

from copy import deepcopy
import six
from nocaselist import NocaseList

from pywbem import CIMClass, CIMQualifierDeclaration, CIMInstance

from pywbem._nocasedict import NocaseDict
from pywbem._utils import _format

from ._baserepository import BaseObjectStore, BaseRepository
from ._utils import _uprint

__all__ = ['InMemoryRepository', 'InMemoryObjectStore']


class InMemoryObjectStore(BaseObjectStore):
    """
    A store for CIM objects of a single type (CIM classes, CIM instances,
    or CIM qualifier declarations) that maintains its data in memory.
    """
    # Documentation for the methods and properties inherited from
    # ~pywbem_mock:`BaseObjectStore` is also inherited in the pywbem
    # documentation. Therefore the methods in this class that are derived
    # from abstrace methods have no documentation string.

    # pylint: disable=line-too-long
    def __init__(self, cim_object_type):

        super(InMemoryObjectStore, self).__init__(cim_object_type)

        self._copy_names = False

        # Define the dictionary that implements the object store.
        # The keys in this dictionary are the names of the objects and
        # the values the corresponding CIM objects.
        if cim_object_type.__name__ in ("CIMClass", 'CIMQualifierDeclaration'):
            self._data = NocaseDict()
        elif cim_object_type.__name__ == 'CIMInstance':
            self._data = {}
            self._copy_names = True
        else:
            assert False, "InMemoryObjectStore: Invalid input parameter {}." \
                .format(cim_object_type)

    def __repr__(self):
        return _format('InMemoryObjectStore(type={0},  dict={1}, size={2}',
                       self._cim_object_type, type(self._data),
                       len(self._data))

    def object_exists(self, name):
        return name in self._data

    def get(self, name, copy=True):
        """
        Get with deepcopy because the pywbem .copy is only middle level and
        we need to completely isolate the repository.
        """
        # pylint: disable=no-else-return
        if name in self._data:
            if copy:
                return deepcopy(self._data[name])
            return self._data[name]
        else:
            raise KeyError('Name {} not in {} object store'
                           .format(name, self._cim_object_type))

    def create(self, name, cim_object):
        assert isinstance(cim_object, self._cim_object_type)

        if name in self._data:
            raise ValueError('Name "{}" already in {} object store'
                             .format(name, self._cim_object_type))
        # Add with deepcopy to completely isolate the copy in the repository
        self._data[name] = deepcopy(cim_object)

    def update(self, name, cim_object):
        assert isinstance(cim_object, self._cim_object_type)

        if name not in self._data:
            raise KeyError('Name "{}" not in {} object store'
                           .format(name, self._cim_object_type))

        # Replace the existing object with a copy of the input object
        self._data[name] = (cim_object)

    def delete(self, name):
        if name in self._data:
            del self._data[name]
        else:
            raise KeyError('Name "{}" not in {} object store'
                           .format(name, self._cim_object_type))

    def iter_names(self):
        """
        Only copies the names for those objects that use CIMNamespaceName
        as the name. The others are immutable ex. classname.
        """
        for name in six.iterkeys(self._data):
            if self._copy_names:
                # Using .copy is sufficient for CIMNamespace name.
                yield(name.copy())
            else:
                yield(name)

    def iter_values(self, copy=True):
        for value in six.itervalues(self._data):
            if copy:
                yield(deepcopy(value))
            else:
                yield(value)

    def len(self):
        return len(self._data)


class InMemoryRepository(BaseRepository):
    """
    A CIM repository that maintains the data in memory using the
    methods defined in its superclass (~pywbem_mock:`BaseObjectStore`).
    This implementation creates the repository as multi-level dictionary
    elements to define the namespaces and within each
    namespace the CIM classes, CIM instances, and CIM qualifiers in the
    repository.
    """
    # Documentation for the methods and properties  isinherited from
    # ~pywbem_mock:`BaseObjectStore` by sphinx when building documentaton.
    # Therefore the methods in this class have no documentation
    # string unless they add or modify documentation in the parent class or
    # are not defined in the parent class. Any method that needs to modifyu
    # the base method documentation must copy the base class documentation.

    def __init__(self, initial_namespace=None):
        """
        Initialize an empty in-memory CIM repository and optionally add a
        namespace in the repository..

        Parameters:

          initial_namespace:(:term:`string` or None):
            Optional initial namespace that will be added to
            the CIM repository.
        """

        # Defines the top level NocaseDict() which defines the
        # namespaces in the repository. The keys of this dictionary
        # are namespace names and the values are dictionaries
        # defining the CIM classes, CIM instances, and CIM qualifier
        # declarations where the keys are "classes", "instances", and
        # "qualifiers" and the value for each is an instance of the
        # class InMemoryObjectStore that containe the CIM objects.
        self._repository = NocaseDict()

        # If an initial namespace is defined, add it to the repository
        if initial_namespace:
            self.add_namespace(initial_namespace)

    def __repr__(self):
        """Display summary of the repository"""
        return _format(
            "InMemoryRepository(data={s._repository})", s=self)

    def print_repository(self, dest=None, ):
        """
        Print the CIM repository to a destination. This displays information on
        the items in the data base and is only a diagnostic tool.

        Parameters:
          dest (:term:`string`):
            File path of an output file. If `None`, the output is written to
            stdout.
        """
        def objstore_info(objstore_name):
            """
            Display the data for the object store
            """
            for ns in self._repository:
                if objstore_name == 'class':
                    store = self.get_class_store(ns)
                elif objstore_name == 'qualifier':
                    store = self.get_qualifier_store(ns)
                else:
                    assert objstore_name == 'instance'
                    store = self.get_instance_store(ns)

                rtn_str = u'Namespace: {} Repo: {} len:{}\n'.format(
                    ns, objstore_name, store.len())
                for val in store.iter_values():
                    rtn_str += (u'{}\n'.format(val))
                return rtn_str

        namespaces = ",".join(self._repository.keys())
        _uprint(dest, _format(u'NAMESPACES: {0}', namespaces))
        _uprint(dest, _format(u'QUALIFIERS: {0}', objstore_info('qualifier')))
        _uprint(dest, _format(u'CLASSES: {0}', objstore_info('class')))
        _uprint(dest, _format(u'INSTANCES: {0}', objstore_info('instance')))

    def validate_namespace(self, namespace):
        if namespace is None:
            raise ValueError("Namespace argument must not be None")

        namespace = namespace.strip('/')
        try:
            self._repository[namespace]
        except KeyError:
            raise KeyError('Namespace "{}" does not exist in repository'.
                           format(namespace))

    def add_namespace(self, namespace):
        if namespace is None:
            raise ValueError("Namespace argument must not be None")

        namespace = namespace.strip('/')

        if namespace in self._repository:
            raise ValueError('Namespace "{}" already in repository'.
                             format(namespace))

        self._repository[namespace] = {}

        # Create the data store for each of the object types.
        self._repository[namespace]['classes'] = InMemoryObjectStore(CIMClass)

        self._repository[namespace]['instances'] = InMemoryObjectStore(
            CIMInstance)

        self._repository[namespace]['qualifiers'] = InMemoryObjectStore(
            CIMQualifierDeclaration)

    def remove_namespace(self, namespace):
        self.validate_namespace(namespace)
        namespace = namespace.strip('/')

        if self.get_class_store(namespace).len() != 0 or \
                self.get_qualifier_store(namespace).len() != 0 or \
                self.get_instance_store(namespace).len() != 0:
            raise ValueError('Namespace {} removal invalid. Namespace not '
                             'empty'.format(namespace))

        del self._repository[namespace]

    @property
    def namespaces(self):
        # pylint: disable=invalid-overridden-method
        # This puts just the dict keys (i.e. namespaces) into the list
        return NocaseList(self._repository)

    def get_class_store(self, namespace):
        if namespace is None:
            raise ValueError("Namespace None not permitted.")
        namespace = namespace.strip('/')
        self.validate_namespace(namespace)
        return self._repository[namespace]['classes']

    def get_instance_store(self, namespace):
        if namespace is None:
            raise ValueError("Namespace None not permitted.")
        namespace = namespace.strip('/')
        self.validate_namespace(namespace)
        return self._repository[namespace]['instances']

    def get_qualifier_store(self, namespace):
        if namespace is None:
            raise ValueError("Namespace None not permitted.")
        namespace = namespace.strip('/')
        self.validate_namespace(namespace)
        return self._repository[namespace]['qualifiers']

    def load(self, other):
        """
        Replace the data in this object with the data from the other object.

        This is used to restore the object from a serialized state, without
        changing its identity.
        """
        # pylint: disable=protected-access
        self._repository = other._repository



#
# (C) Copyright 2020 InovaDevelopment.com
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Karl  Schopmeyer <inovadevelopment.com>
#

"""
This module implements a user-defined provider for the class CIM_Namespace

This  provider manages creation and deletion of the instances of the
CIM_Namespace class and limits that activity to the server environment interop
classname.  The CIM_Namespace class is a DMTF defined class that
defines an instance for each namespace in the server environment and allows
the creation and deletion of server CIM namespaces to be controlled through
creation and deletion of instances of  the CIM_Namespace class.
"""

from nocaselist import NocaseList

from pywbem import CIMError, CIMInstance, CIM_ERR_NOT_SUPPORTED, \
    CIM_ERR_INVALID_PARAMETER
from pywbem._nocasedict import NocaseDict
from pywbem._utils import _format

# TODO: Should be able to import from pywbem_mock
from ._instancewriteprovider import InstanceWriteProvider

from .config import OBJECTMANAGERNAME, SYSTEMNAME, SYSTEMCREATIONCLASSNAME, \
    OBJECTMANAGERCREATIONCLASSNAME


# CIM class name of the namespace class implemented in this provider
NAMESPACE_CLASSNAME = 'CIM_Namespace'


class CIMNamespaceProvider(InstanceWriteProvider):
    # pylint: disable=line-too-long
    """
    Implements the user defined provider for the class CIM_Namespace.

    This provider provides the create, modify, and delete methods for
    adding an instance of the class CIM_Namspace when a namespace is created or
    deleted in a pywbem mock environment.

    This class and the instances of this class only exist in the WBEM server
    Interop namespace per DMTF definitions.

    This class __init__ saves the  cimrepository variable used by methods in
    this class and by methods of its superclasses (i.e.  InstanceWriteProvider).

    The provider defines the class level attribute `provider_classnames` (CIM_Namespace)

    This provider presumes that an Interop namespace has been created before
    the provider object is constructed and fails the constructor if
    there is not interop_namespace
    """  # noqa: E501
    # pylint: enable=line-too-long

    # This class level attribute must exist to define the CIM classname(es) .
    # for which this provider is responsible.
    #: provider_classnames (:term:`string`):
    #:        The classname for this provider
    provider_classnames = NAMESPACE_CLASSNAME

    def __init__(self, cimrepository):
        """
        Parameters:

          cimrepository (:class:`~pywbem_mock.BaseRepository` or subclass):
            Defines the CIM repository to be used by the provider.
        """
        super(CIMNamespaceProvider, self).__init__(cimrepository)

        if not self.find_interop_namespace():
            raise CIMError(
                CIM_ERR_INVALID_PARAMETER,
                _format("Cannot create namespace provider (for class {0}): "
                        "No Interop namespace exists in the CIM repository. "
                        "Valid Interop namespaces are: {1}",
                        NAMESPACE_CLASSNAME,
                        ", ".join(self.interop_namespace_names)))

        self.installed = False  # test if provider previously installed.

    def __repr__(self):
        return _format(
            "CIMNamespaceProvider("
            "provider_type={s.provider_type}, "
            "provider_classnames={s.provider_classnames})",
            s=self)

    def CreateInstance(self, namespace, new_instance):
        """
        Create an instance of the CIM_Namespace class in an Interop namespace
        of the CIM repository, and if not yet existing create the new namespace
        in the CIM repository.

        See `~pywbem_mock.InstanceWriteProvider.CreateInstance` for
        documentation of validation and description of input parameters, noting
        extra conditions for this provider as described below:

        Parameters:

          namespace (:term:`string`):
            Must be a valid Interop namespace.

          new_instance (:class:`~pywbem.CIMInstance`):
            The following applies regarding its properties:
            * 'Name' property: This property is required since it defines the
              name of the new namespace to be created.
            * 'CreationClassName' property: This property is required and its
              value must match the class name of the new instance.

        Raises:

          :exc:`~pywbem.CIMError`: (CIM_ERR_INVALID_PARAMETER) if namespace
            is not the Interop namespace in the CIM repository or the Name
            property does not exist or the other properties cannot be added to
            the instance.
        """

        if not self.is_interop_namespace(namespace):
            raise CIMError(
                CIM_ERR_INVALID_PARAMETER,
                _format(
                    "Cannot create instance of class {0!A} in namespace {1!A}: "
                    "The namespace is not an Interop namespace. "
                    "Valid Interop namespaces are: {2!A}",
                    new_instance.classname, namespace,
                    ", ".join(self.interop_namespace_names)))

        ccn_pname = 'CreationClassName'
        name_pname = 'Name'

        # Validate that required properties are specified in the new instance
        for pn in [name_pname, ccn_pname]:
            if pn not in new_instance:
                raise CIMError(
                    CIM_ERR_INVALID_PARAMETER,
                    _format(
                        "Cannot create instance of class {0!A} in "
                        "namespace {1!A}: "
                        "Missing required property {2!A} in new_instance",
                        new_instance.classname, namespace, pn))

        # Get and normalize the new namespace name
        new_namespace = new_instance[name_pname]  # Property value
        new_namespace = new_namespace.strip('/')

        # Write it back to the instance in case it was changed
        new_instance[name_pname] = new_namespace

        # Validate other property values in the new instance
        if new_instance[ccn_pname].lower() != new_instance.classname.lower():
            raise CIMError(
                CIM_ERR_INVALID_PARAMETER,
                _format(
                    "Cannot create instance of class {0!A} in namespace "
                    "{1!A}: Value of property {2|A} in new_instance does not "
                    "match class name but is {3!A}",
                    new_instance.classname, namespace,
                    ccn_pname, new_instance[ccn_pname]))

        # Create the new namespace in the CIM repository, if needed.
        if new_namespace not in self.cimrepository.namespaces:
            self.cimrepository.add_namespace(new_namespace)

        # Create the CIM instance for the new namespace in the CIM repository,
        # by delegating to the default provider method.
        return super(CIMNamespaceProvider, self).CreateInstance(
            namespace, new_instance)

    def ModifyInstance(self, modified_instance, IncludeQualifiers=None):
        """
        Modification of CIM_Namespace instance not allowed
        """
        raise CIMError(
            CIM_ERR_NOT_SUPPORTED,
            _format("Modification of {0} instances is not allowed: "
                    "{1!A}",
                    NAMESPACE_CLASSNAME, modified_instance.path))

    def DeleteInstance(self, InstanceName):
        """
        Delete an instance of the CIM_Namespace class in an Interop namespace
        of the CIM repository, and in addition delete the namespace represented
        by it in the CIM repository.

        See `~pywbem_mock.InstanceWriteProvider.CreateInstance` for
        documentation of validation and description of input parameters, noting
        extra conditions for this provider as described below:

        The namespace to be deleted must be empty and must not be the Interop
        namespace.

        Parameters:

          InstanceName: (:class:`~pywbem.CIMInstanceName`):
            The keybinding `Name` must exist; it defines the namespace to be
            deleted.

        Raises:

          :exc:`~pywbem.CIMError`: (CIM_ERR_INVALID_PARAMETER)
          :exc:`~pywbem.CIMError`: (CIM_ERR_NAMESPACE_NOT_EMPTY)
        """

        # The provider dispatcher ensures that provider methods are only called
        # for the registered classes.
        # And this provider sets only a single class, not a list.
        assert InstanceName.classname.lower() == \
            CIMNamespaceProvider.provider_classnames.lower()

        remove_namespace = InstanceName.keybindings['Name']

        if self.is_interop_namespace(remove_namespace):
            raise CIMError(
                CIM_ERR_INVALID_PARAMETER,
                _format("Cannot delete instance {0!A} from the CIM repository: "
                        "This instance represents the Interop namespace {1!A} "
                        "which must not be deleted.",
                        InstanceName, remove_namespace))

        # Delete the namespace from the CIM repository.
        # This call verifies that the namespace is empty and raises
        # CIMError(CIM_ERR_NAMESPACE_NOT_EMPTY) if not empty.
        self.remove_namespace(remove_namespace)

        # Delete the instance from the CIM repository
        instance_store = self.cimrepository.get_instance_store(
            InstanceName.namespace)
        instance_store.delete(InstanceName)

    def post_register_setup(self, conn):
        """
        Method called by FakedWBEMConnection.register_provider to complete
        initialization of this provider.  This method is called after
        the required classes are installed in the cim_repository

        This is necessary because pywbem_mock does not allow user-defined
        providers for the instance read operations such as EnumerateInstances
        so the instance for each namespace must be exist in the repository.

        This method:

        1. Inserts instances of CIM_Namespace for every namespace in the
           CIM repository.
        """
        assert self.installed is False

        self.installed = True

        # Temp code to validate class installed.  Will generate exception
        # if class does not exist.
        interop_namespace = conn.find_interop_namespace()
        provider_classname = self.provider_classnames
        conn.GetClass(provider_classname, interop_namespace)

        # Create instance CIM_Namespace for each existing namespace

        insts = conn.EnumerateInstances('CIM_Namespace',
                                        namespace=conn.find_interop_namespace(),
                                        LocalOnly=False,
                                        DeepInheritance=False)

        # List of namespaces for which there are CIM instances
        inst_ns_list = NocaseList([inst.name for inst in insts])

        klass = conn.GetClass(provider_classname, interop_namespace,
                              LocalOnly=False,
                              IncludeQualifiers=True,
                              IncludeClassOrigin=True)

        # Set up any CIM_Namespace instances that are still missing given the
        # namespaces that exist in the CIM repository.
        for ns in conn.cimrepository.namespaces:
            if ns not in inst_ns_list:
                self.create_cimnamespace_instance(conn, ns, interop_namespace,
                                                  klass)

    @staticmethod
    def create_cimnamespace_instance(conn, namespace, interop_namespace, klass):
        """
        Build and execute CreateInstance for an instance of CIM_Namespace using
        the `namespace` parameter as the value of the name property in the
        instance. The instance is created in the `interop_namespace`

        Parameters:

          namespace (:term:`string`):
            Namespace that this instance defines.

          interop_namespace (:term:`string`):
            Interop namespace for this environment.

          klass (:class:`CIM_Namespace`):
            The CIM class CIM_Namespace which is used to create the instance.

        Raises:

            :exc:`~pywbem.CIMError`: For errors encountered with CreateInstance
        """
        properties = NocaseDict(
            [('Name', namespace),
             ('CreationClassName', klass.classname),
             ('ObjectManagerName', OBJECTMANAGERNAME),
             ('ObjectManagerCreationClassName', OBJECTMANAGERCREATIONCLASSNAME),
             ('SystemName', SYSTEMNAME),
             ('SystemCreationClassName', SYSTEMCREATIONCLASSNAME)])

        new_instance = CIMInstance.from_class(klass, property_values=properties)
        conn.CreateInstance(new_instance, interop_namespace)



"""
Test memory leaks for dictionary classes used by pywbem.
"""

from __future__ import absolute_import, print_function

import sys
from collections import OrderedDict
try:
    from django.utils.datastructures import SortedDict as Django_SortedDict
except ImportError:
    Django_SortedDict = None
import pytest
import yagot

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ..utils import import_installed
pywbem = import_installed('pywbem')
from pywbem._cim_obj import NocaseDict  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


# collections.OrderedDict has fixed its ref.cycles starting with Python 3.2
ORDEREDDICT_LEAKFREE_VERSION = (3, 2)


@pytest.mark.xfail(
    sys.version_info < ORDEREDDICT_LEAKFREE_VERSION,
    reason="OrderedDict has reference cycles on py<3.2")
@yagot.garbage_checked()
def test_leaks_OrderedDict_empty():
    """
    Test function with empty OrderedDict object.

    Note: collections.OrderedDict has memory leaks on Python 2.7; see
    https://bugs.python.org/issue9825. That issue was fixed in Python 3.2, but
    the change in Python 2.7 apparently was not sufficient to remove the leak.
    """
    _ = OrderedDict()


@pytest.mark.skipif(
    Django_SortedDict is None,
    reason="django with SortedDict is not installed")
@yagot.garbage_checked()
def test_leaks_Django_SortedDict_empty():
    """
    Test function with empty django SortedDict object.

    Note: Django's SortedDict has been removed from the django package as of
    version 1.9. This test requires django<1.9 to be installed and is skipped
    otherwise.
    """
    _ = Django_SortedDict()


@pytest.mark.xfail(
    sys.version_info < ORDEREDDICT_LEAKFREE_VERSION,
    reason="NocaseDict uses OrderedDict")
@yagot.garbage_checked()
def test_leaks_NocaseDict_empty():
    """
    Test function with empty NocaseDict object.
    """
    _ = NocaseDict()


@pytest.mark.xfail(
    sys.version_info < ORDEREDDICT_LEAKFREE_VERSION,
    reason="NocaseDict uses OrderedDict")
@yagot.garbage_checked()
def test_leaks_NocaseDict_one():
    """
    Test function with NocaseDict object containing one string item.
    """
    _ = NocaseDict(a='b')
